"""
Command-line script: run an analysis module

Loads a class from a module (import path or file),
constructs an instance with the remaining arguments to parse,
and runs it
"""

import argparse
import importlib.util
import logging
import os.path


def main():
    parser = argparse.ArgumentParser(description="Run an analysis module", add_help=False)
    parser.add_argument("-m", "--module", type=str,
                        default="bamboo.analysismodules:AnalysisModule",
                        help="Module to run (format: modulenameOrPath[:classname])")
    parser.add_argument("--help", "-h", action="store_true",
                        help="Print this help message and exit")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Run in verbose mode")
    args, remArgs = parser.parse_known_args()
    # pass arguments on
    remArgs.append(f"--module={args.module}")
    if args.help:
        remArgs.append("--help")
    if args.verbose:
        remArgs.append("--verbose")

    logModules = [f"bamboo.{bMod}" for bMod in (
        "analysismodules", "analysisutils",
        "batch", "batch_slurm", "batch_htcondor",
        "dataframebackend", "treeoperations", "treedecorators", "plots")]
    logLevel = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(level=logLevel)  # FIXME specify a handler
    for logMod in logModules:
        modLogger = logging.getLogger(logMod)
        modLogger.setLevel(logLevel)

    modArg = args.module
    clName = None
    if ":" in modArg:
        if modArg.count(":") > 1:
            raise RuntimeError(f"More than one : in module argument '{modArg}'")
        modArg, clName = tuple(modArg.split(":"))
    # modArg should now be a python module name or path now
    spec = None
    try:
        spec = importlib.util.find_spec(modArg)
    except Exception:
        pass
    if spec:  # is a module
        if not clName:
            clName = spec.split(".")[-1]
    elif os.path.exists(modArg):
        modName = os.path.splitext(os.path.basename(modArg))[0]
        if not clName:
            clName = modName
        spec = importlib.util.spec_from_file_location(modName, modArg)
    else:
        raise RuntimeError(
            f"Module argument '{modArg}' is neither "
            "an importable module or an existing path")
    # do the import
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    if not hasattr(mod, clName):
        raise RuntimeError(
            "Module {0} has no class with name {1}, "
            "please specify it using --module=modulenameOrPath:classname")
    modCls = getattr(mod, clName)
    modInst = modCls(remArgs)
    modInst.run()



# A first attempt at `ROOT::RDataFrame` wrappers in python
# ========================================================
#
# *Warning*: this is very much work in progress... the idea is to wrap a NanoAOD in an analyst-friendly way,
# to make writing analysis code (from tree to plots) as flexible as possible,
# such that python scripting can be used to avoid repetition and keep the code compact and maintainable.
#
# A few imports to get started, and debug logging such that we can see what is going on.
import logging
logging.basicConfig(level=logging.DEBUG)
from bamboo.treedecorators import decorateNanoAOD, NanoAODDescription  # noqa: E402
from bamboo.dataframebackend import DataframeBackend  # noqa: E402
from bamboo.analysisutils import addLumiMask  # noqa: E402
import ROOT  # noqa: E402
if hasattr(ROOT, "enableJSVis"):
    ROOT.enableJSVis()  # for inline plots in the notebook
# `decorateNanoAOD` will wrap the NanoAOD in a more pythonic set of objects, `DataframeBackend` takes care of
# passing the right instructions to a `ROOT::RDataFrame` instance when we define plots.
f = ROOT.TFile.Open("NanoAOD_SingleMu_test.root")
t = decorateNanoAOD(f.Get("Events"), NanoAODDescription.get("v5", year="2016", isMC=True))
be, noSel = DataframeBackend.create(t)
noSel = addLumiMask(
    noSel, "Cert_271036-284044_13TeV_23Sep2016ReReco_Collisions16_JSON.txt",
    runRange=(276831, 277420), runAndLS=(t.run, t.luminosityBlock)
)
# At this point we have a wrapped tree `t`, a reference to the backend
# (which we will only need in the end, to retrieve the plot results),
# and a basic selection `noSel`, which lets all events pass.
# To define a plot, we will need to pass (at the very least) a name for it,
# a variable to plot (with a suitable binning), and a selection to apply
# (which could be noSel if the plot is to be filled for all records).
from bamboo.plots import Plot, EquidistantBinning  # noqa: E402
import bamboo.treefunctions as op  # noqa: E402
# The `Plot` class is defined in the plots module, together with helper classes
# (e.g. binnings, such as `EquidistantBinning` and `VariableBinning`).
# The `treefunctions` module defines mathematical functions etc. for the
# tree proxy objects.

# Let's start with an example to demonsrate the power of the plot wrappers:
# a dimuon mass plot. First we'll need a dimuon selection
twoMuSel = noSel.refine("twoMuons", cut=[op.rng_len(t.Muon) > 1])
# Then we can define the plot, using the momenta of the two leading muons.
# Note that the `p4` branch (or `Muon_p4`) is not in the original tree.
# It is transparently created on-the-fly from the pt, eta, phi and mass
# branches (the backend will tell the `RDataFrame` to construct it when needed).
myMPlot = Plot.make1D(
    "dimu_M", op.invariant_mass(t.Muon[0].p4, t.Muon[1].p4), twoMuSel,
    EquidistantBinning(100, 20., 120.), title="Dimuon invariant mass"
)
# The original pt attribute of the muon (from `Muon_pt`) is als still there:
myPlot = Plot.make1D(
    "dimu_Pt2", t.Muon[1].pt, twoMuSel,
    EquidistantBinning(50, 0., 200.), title="Trailing muon PT"
)
myOtherPlot = Plot.make1D(
    "dimu_Pt1", t.Muon[0].p4.Pt(), twoMuSel,
    EquidistantBinning(50, 0., 200.), title="Leading muon PT"
)

# After this warmup, let's get a bit more realistic, and define a selected
# muon collection and plot the pT of the first one (if there are).
myMu = op.select(t.Muon, lambda mu: op.abs(mu.eta) < 2.)
myNsPlot = Plot.make1D(
    "nMu10", op.rng_len(myMu), noSel,
    EquidistantBinning(10, 0., 10.), title="Number of muons with p_{T} > 10 GeV"
)
hasMyMuSel = noSel.refine("oneMyMuon", cut=[op.rng_len(myMu) > 0])
leadMyMuPtPlot = Plot.make1D(
    "leadMymuPt", myMu[0].pt, hasMyMuSel,
    EquidistantBinning(50, 0., 200.), title="Leading myMuon PT"
)


def isVeryCentral(mu):
    return op.abs(mu.eta) < 2.


hasCentralMuon = noSel.refine("hasCentralMu", cut=[op.rng_any(t.Muon, isVeryCentral)])
leadCentralMu = op.rng_find(t.Muon, isVeryCentral)
centralMuPtPlot = Plot.make1D(
    "centMu_Pt1", leadCentralMu.pt, hasCentralMuon,
    EquidistantBinning(50, 0., 200.), title="Leading central Muon PT"
)
# Range functions that (unlike `rng_find` and `rng_any`) always process
# the complete range are also defined, let's for example sum the pseudorapidities
# of the very central, and then all muons (despite the limited physical interest):
myMuSumEtaPlot = Plot.make1D(
    "sumMyMuEta", op.rng_sum(myMu, lambda mu: mu.eta), noSel,
    EquidistantBinning(50, -10., 10.), title="Sum of myMu eta"
)
muSumEtaPlot = Plot.make1D(
    "sumMuEta", op.rng_sum(t.Muon, lambda mu: mu.eta), noSel,
    EquidistantBinning(50, -10., 10.), title="Sum of mu eta"
)
# A very useful function looks for the object with the largest/smallest
# value for some function, `rng_max`/`min_element`, e.g. the most central muon:
mostCentralMuon = op.rng_min_element_by(t.Muon, lambda mu: op.abs(mu.eta))
mostCentralMuPt = Plot.make1D(
    "mostCentralMuPt", mostCentralMuon.pt, hasCentralMuon,
    EquidistantBinning(50, 0., 200.), title="Most central muon PT"
)

# Also scalefactors are easy to add, as weight factors for the Selection:
from bamboo.scalefactors import get_scalefactor  # noqa: E402
lSF = get_scalefactor("lepton", ("muon_2016_80", "iso_loose_id_loose"), combine="weight")
twoMuSelSF = noSel.refine("twoMuonsSF", cut=[op.rng_len(t.Muon) > 1],
                          weight=[lSF(t.Muon[0], withMCCheck=False),
                                  lSF(t.Muon[1], withMCCheck=False)])
newMPlot = Plot.make1D(
    "dimuSF_M", op.invariant_mass(t.Muon[0].p4, t.Muon[1].p4), twoMuSelSF,
    EquidistantBinning(100, 20., 120.), title="Dimuon invariant mass"
)


# The basic infrastructure for on-the-fly jet systematic variations is also in place,
# this example does not modify the jet branches, but could be made to do so by calling
# a method in C++. `t.Jet.nominal` (or `t.Jet["nominal"]`) has the original jet,
# `t.Jet[variation]` a modified jet container.
op.addKinematicVariation(t.Jet, "no-op", modif=(lambda j: j.p4), pred=(lambda mp4, j: mp4.Pt() > 10.))
nJets = Plot.make1D(
    "nJets", op.rng_len(t.Jet["no-op"]), noSel,
    EquidistantBinning(10, 0., 10.), title="Unmodified jet multiplicity")

# Making two-particle combinations (higher N to be added) is also easy:
osdimu = op.combine(t.Muon, N=2, pred=lambda m1, m2: m1.charge != m2.charge)
oselmu = op.combine((t.Electron, t.Muon), pred=lambda el, mu: el.charge != mu.charge)
nOSDimu = Plot.make1D(
    "nOsDimu", op.rng_len(osdimu), noSel, EquidistantBinning(10, 0., 10.),
    title="Number of opposite-sign dimuon pairs")
nOSelmu = Plot.make1D(
    "nOselmu", op.rng_len(oselmu), noSel,
    EquidistantBinning(10, 0., 10.),
    title="Number of opposite-sign electron-muon pairs")
# TODO stress-test: use selected muons, select afterwards, combine electrons and muons

tagEle = op.select(
    t.Electron,
    lambda el: op.AND(
        el.pt >= 30.,
        op.abs(el.eta + el.deltaEtaSC) < 2.1,
        op.NOT(op.in_range(1.4442, op.abs(el.eta + el.deltaEtaSC), 1.566)),
        op.rng_any(
            t.TrigObj,
            lambda to: op.AND(
                to.id == 11,
                to.filterBits & 2,
                op.sqrt((to.eta - el.eta)**2 + (to.phi - el.phi)**2) < .4))
    )
)  # TODO review 'filterBits'
nTagEle = Plot.make1D("nTagEle", op.rng_len(tagEle), noSel, EquidistantBinning(10, 0., 10.),
                      title="Number of tag electrons (partial selection)")
probeEle = op.select(t.Electron, lambda el: op.abs(el.eta + el.deltaEtaSC) < 2.5)
eleTagAndProbe = op.combine(
    (tagEle, probeEle),
    pred=lambda elt, elp: op.in_range(50., op.invariant_mass(elt.p4, elp.p4), 130.),
    samePred=lambda o1, o2: o1.idx != o2.idx)
# TODO add trigger event selection
hasTagAndProbeEle = noSel.refine("hasTagAndProbe", cut=(op.rng_len(eleTagAndProbe) > 0))
theTnPpair = eleTagAndProbe[0]  # TODO add more strategies
tnpelemass = Plot.make1D(
    "eleTnPmassInteg", op.invariant_mass(theTnPpair[0].p4, theTnPpair[1].p4),
    hasTagAndProbeEle, EquidistantBinning(80, 50., 130.),
    title="Tag&Probe electron pair invariant mass")

# That's a nice start - but we haven't plotted any histograms yet!
# The `ROOT::RDataFrame` will do that as soon as we ask for one of them:
cv = ROOT.TCanvas("c1", "A few dimuon plots", 1500, 350)
cv.Divide(3)
cv.cd(1)
be.getResults(myPlot)[0].Draw()
cv.cd(2)
be.getResults(myOtherPlot)[0].Draw()
cv.cd(3)
be.getResults(myMPlot)[0].Draw()
cv.Update()
cv.Draw()

cv2 = ROOT.TCanvas("c2")
cv2.Divide(3)
cv2.cd(1)
be.getResults(myNsPlot)[0].Draw()
cv2.cd(2)
be.getResults(leadMyMuPtPlot)[0].Draw()
cv2.cd(3)
be.getResults(centralMuPtPlot)[0].Draw()
cv2.Update()
cv2.Draw()

cv3 = ROOT.TCanvas("c3")
cv3.Divide(2)
cv3.cd(1)
be.getResults(myMuSumEtaPlot)[0].Draw()
cv3.cd(2)
be.getResults(mostCentralMuPt)[0].Draw()
cv3.Update()
cv3.Draw()

cv4 = ROOT.TCanvas("c4")
cv4.Divide(2)
cv4.cd(1)
be.getResults(newMPlot)[0].Draw()
cv4.cd(2)
be.getResults(nJets)[0].Draw()
cv4.Update()

cv5 = ROOT.TCanvas("c5")
cv5.Divide(2)
cv5.cd(1)
be.getResults(nOSDimu)[0].Draw()
cv5.cd(2)
be.getResults(nOSelmu)[0].Draw()
cv5.Update()

cv6 = ROOT.TCanvas("c6")
cv6.Divide(2)
cv6.cd(1)
be.getResults(nTagEle)[0].Draw()
cv6.cd(1)
be.getResults(tnpelemass)[0].Draw()
cv6.Update()

# There are a few more things to figure out:
#   - integrate with slurm. The easiest is probably to have a common module
#     that runs either in a single thread or on slurm, that gets a list of plots
#     from a method `definePlots(tree, noSel)`, and saves all histograms to a file.
#   - Combinatorics (pairs and triplets of objects)
#   - "skims" (should be supported through the `Snapshot` method)
#   - other tree formats?
#   - `your idea here`



import os
import os.path

import pytest

# check if libtensorflow can be found
# (duplicates loadTensorflowC, but cannot just try to load because libtorch should be loaded first)
from bamboo.root import findLibrary
from helpers import isclose_float

tfc_found = findLibrary("libtensorflow")
if not tfc_found:
    if "VIRTUAL_ENV" in os.environ:
        tfclib_guessed_path = os.path.join(
            os.environ["VIRTUAL_ENV"], "lib", "libtensorflow.so")
        if os.path.exists(tfclib_guessed_path):
            tfc_found = tfclib_guessed_path
pytestmark = pytest.mark.skipif(not tfc_found, reason="Tensorflow-C not found")

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def tfc_nn1():
    from bamboo.root import gbl, loadTensorflowC
    loadTensorflowC()

    def vstring(*args):
        res = gbl.std.vector["std::string"]()
        for arg in args:
            res.push_back(arg)
        return res

    yield gbl.bamboo.TensorflowCEvaluator(
        os.path.join(testData, "nn1.pb"),
        vstring("dense_1_input"),
        vstring("dense_3/Softmax"))


def test_tfc_load(tfc_nn1):
    assert tfc_nn1


def test_tfc_evalZero(tfc_nn1):
    from bamboo.root import gbl
    v_in = gbl.std.vector["float"]()
    for _i in range(5):
        v_in.push_back(.7)
    v_out = tfc_nn1.evaluate(v_in)
    assert (v_out.size() == 2) and isclose_float(sum(v_out), 1.)


def test_tfc_evalRandom(tfc_nn1):
    from bamboo.root import gbl
    import random
    v_in = gbl.std.vector["float"]()
    for _i in range(5):
        v_in.push_back(random.uniform(0., 1.))
    v_out = tfc_nn1.evaluate(v_in)
    assert (v_out.size() == 2) and isclose_float(sum(v_out), 1.)



import pytest


def pytest_addoption(parser):
    parser.addoption("--plots-reference", action="store", help="Directory with reference plots")
    parser.addoption("--plots-output", action="store", help="Directory to store the produced plots")
    parser.addoption("--with-compiled", action="store_true",
                     help="Run the tests with the compiled backend")


@pytest.fixture(scope="session")
def plots_reference(request):
    return request.config.getoption("--plots-reference")


@pytest.fixture(scope="session")
def plots_output(request):
    return request.config.getoption("--plots-output")


@pytest.fixture(scope="session")
def with_compiled(request):
    return request.config.getoption("--with-compiled")


def pytest_configure(config):
    config.addinivalue_line("markers", "plotswithreference: produce and compare plots (slow)")
    config.addinivalue_line("markers", "withcompiled: produce and compare plots (slow, high memory)")


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--plots-reference"):  # --plots-reference not given in cli: skip plots
        skip_plots_with_reference = pytest.mark.skip(reason="need --plots-reference option to run")
        for item in items:
            if "plotswithreference" in item.keywords:
                item.add_marker(skip_plots_with_reference)
    if not config.getoption("--with-compiled"):  # --with-compiled not given in cli: skip plots
        skip_with_compiled = pytest.mark.skip(reason="need --with-compiled option to run")
        for item in items:
            if "withcompiled" in item.keywords:
                item.add_marker(skip_with_compiled)



import os.path

import pytest

from bamboo.root import findLibrary
from helpers import isclose_float

# check if lwtnn is found (the bridge library should be available then)
lwtnn_found = findLibrary("liblwtnn")
if not lwtnn_found:
    if "VIRTUAL_ENV" in os.environ:
        lwtnnlib_guessed_path = os.path.join(os.environ["VIRTUAL_ENV"], "lib", "liblwtnn.so")
        if os.path.exists(lwtnnlib_guessed_path):
            lwtnn_found = lwtnnlib_guessed_path
pytestmark = pytest.mark.skipif(not lwtnn_found, reason="lwtnn library not found")

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def lwtnn_nn1():
    from bamboo.root import loadlwtnn, gbl
    loadlwtnn()
    inputs = gbl.std.vector["std::pair<std::string,std::string>"]()
    input_item = gbl.std.pair["std::string,std::string"]
    for i in range(5):
        inputs.push_back(input_item("node_0", f"variable_{i:d}"))
    outputs = gbl.std.vector["std::string"]()
    for i in range(2):
        outputs.push_back(f"out_{i:d}")
    yield gbl.bamboo.LwtnnEvaluator(os.path.join(testData, "nn1_lwtnn.json"), inputs, outputs)


def test_lwtnn_load(lwtnn_nn1):
    assert lwtnn_nn1


def test_lwtnn_evalZero(lwtnn_nn1):
    from bamboo.root import gbl
    v_in = gbl.std.vector["float"]()
    for _i in range(5):
        v_in.push_back(.7)
    v_out = lwtnn_nn1.evaluate(v_in)
    assert (v_out.size() == 2) and isclose_float(sum(v_out), 1.)


def test_lwtnn_evalRandom(lwtnn_nn1):
    from bamboo.root import gbl
    import random
    v_in = gbl.std.vector["float"]()
    for _i in range(5):
        v_in.push_back(random.uniform(0., 1.))
    v_out = lwtnn_nn1.evaluate(v_in)
    assert (v_out.size() == 2) and isclose_float(sum(v_out), 1.)



import os.path

import pytest

from helpers import isclose_float

testData = os.path.join(os.path.dirname(__file__), "data")


def makeParameters(**kwargs):
    """ Construct Parameters argument for BinnedValues """
    from bamboo.root import gbl, loadBambooExtensions
    loadBambooExtensions()
    gbl.BinningVariable  # somehow loads Pt, Eta etc.
    params = gbl.Parameters()
    for k, v in kwargs.items():
        params.set(getattr(gbl, k), v)
    return params


@pytest.fixture(scope="module")
def sf_leptonSingle():
    from bamboo.root import gbl, loadBambooExtensions
    loadBambooExtensions()
    gbl.SystVariation  # somehow loads Nominal etc.
    elSFJSON = os.path.join(testData, "Electron_EGamma_SF2D_loose_moriond17.json")
    return gbl.ScaleFactor(elSFJSON)


@pytest.fixture(scope="module")
def sf_btagSingle():
    from bamboo.root import gbl, loadBambooExtensions
    loadBambooExtensions()
    gbl.SystVariation  # somehow loads Nominal etc.
    lightJSON = os.path.join(testData, "BTagging_loose_lightjets_incl_DeepJet_2016Legacy.json")
    cjetJSON = os.path.join(testData, "BTagging_loose_cjets_comb_DeepJet_2016Legacy.json")
    bjetJSON = os.path.join(testData, "BTagging_loose_bjets_comb_DeepJet_2016Legacy.json")
    return gbl.BTaggingScaleFactor(lightJSON, cjetJSON, bjetJSON)


@pytest.fixture(scope="module")
def puWeight():
    from bamboo.root import gbl, loadBambooExtensions
    loadBambooExtensions()
    puWeightJSON = os.path.join(os.path.dirname(__file__), "data", "puweights.json")
    return gbl.ScaleFactor(puWeightJSON)


@pytest.fixture(scope="module")
def test_lepSingle_constructEval(sf_leptonSingle):
    from bamboo.root import gbl
    central = 0.9901639223098755
    error = 0.19609383660010074
    assert isclose_float(
        sf_leptonSingle.get(makeParameters(Pt=20., Eta=1.5), gbl.Nominal),
        central)
    assert isclose_float(
        sf_leptonSingle.get(makeParameters(Pt=20., Eta=1.5), gbl.Up),
        central + error)
    assert isclose_float(
        sf_leptonSingle.get(makeParameters(Pt=20., Eta=1.5), gbl.Down),
        central - error)


def test_btagSingle_constructEval(sf_btagSingle):
    from bamboo.root import gbl
    central = 1.0912339687347412
    error = 0.057828545570373535
    assert isclose_float(
        sf_btagSingle.get(
            makeParameters(Pt=50., Eta=1.5, BTagDiscri=.5),
            gbl.IJetScaleFactor.Light, gbl.Nominal),
        central)
    assert isclose_float(
        sf_btagSingle.get(
            makeParameters(Pt=50., Eta=1.5, BTagDiscri=.5),
            gbl.IJetScaleFactor.Light, gbl.Up),
        central + error)
    assert isclose_float(
        sf_btagSingle.get(
            makeParameters(Pt=50., Eta=1.5, BTagDiscri=.5),
            gbl.IJetScaleFactor.Light, gbl.Down),
        central - error)


def test_puWeight_constructEvalInRange(puWeight):
    from bamboo.root import gbl
    assert isclose_float(
        puWeight.get(makeParameters(NumTrueInteractions=20.5), gbl.Nominal),
        1.0656023337493363)


def test_puWeight_evalBinEdge(puWeight):
    from bamboo.root import gbl
    assert isclose_float(
        puWeight.get(makeParameters(NumTrueInteractions=20.), gbl.Nominal),
        1.0656023337493363)


def test_puWeight_evalOutOfRangeBelow(puWeight):
    from bamboo.root import gbl
    assert isclose_float(
        puWeight.get(makeParameters(NumTrueInteractions=-.5), gbl.Nominal),
        0.36607730074755906)


def test_puWeight_evalOutOfRangeAbove(puWeight):
    from bamboo.root import gbl
    assert isclose_float(
        puWeight.get(makeParameters(NumTrueInteractions=100.), gbl.Nominal),
        0.001723281482149061)



""" Example CMS Phase2 simulation analysis module """

###########################################
#                                         #
# methods and classes for a common module #
#                                         #
###########################################

from bamboo.analysismodules import AnalysisModule, HistogramsModule


class CMSPhase2SimModule(AnalysisModule):
    """ Base module for processing Phase2 flat trees """
    def __init__(self, args):
        super().__init__(args)

    def prepareTree(self, tree, sample=None, sampleCfg=None):
        from bamboo.treedecorators import decorateCMSPhase2SimTree
        from bamboo.dataframebackend import DataframeBackend
        t = decorateCMSPhase2SimTree(tree, isMC=True)
        be, noSel = DataframeBackend.create(t)
        return t, noSel, be, tuple()

    def mergeCounters(self, outF, infileNames, sample=None):
        from bamboo.root import gbl
        outF.cd()
        hNEvts = gbl.TH1F("nEvents", "Number of events", 1, 0., 1.)
        for fn in infileNames:
            f = gbl.TFile.Open(fn)
            tup = f.Get("myana/mytree")  # FIXME
            if not tup:
                raise RuntimeError(f"File {fn} does not have a tree {self.args.treeName}")
            hNEvts.Fill(0, tup.GetEntries())
        outF.cd()
        hNEvts.Write("nEvents")

    def readCounters(self, resultsFile):
        hNEvts = resultsFile.Get("nEvents")
        return {"nEvents": hNEvts.GetBinContent(1)}


class CMSPhase2SimHistoModule(CMSPhase2SimModule, HistogramsModule):
    """ Base module for producing plots from Phase2 flat trees """
    def __init__(self, args):
        super().__init__(args)


##############################
#                            #
# The actual analysis module #
#                            #
##############################

class CMSPhase2SimTest(CMSPhase2SimHistoModule):
    """ Example plotter module for Phase2 flat trees """
    def definePlots(self, t, noSel, sample=None, sampleCfg=None):
        from bamboo.plots import Plot
        from bamboo.plots import EquidistantBinning as EqB
        from bamboo import treefunctions as op

        plots = []

        electrons = op.select(t.elec, lambda el: op.AND(
            el.pt > 20.,
            op.abs(el.eta) < 2.5,
            el.idpass, el.isopass))
        muons = op.select(t.muon, lambda mu: op.AND(
            mu.pt > 20.,
            op.abs(mu.eta) < 2.4,
            mu.idpass, mu.isopass))
        cleanedJets = op.select(t.jet, lambda j: op.AND(
            op.NOT(op.rng_any(electrons, lambda el: op.deltaR(el.p4, j.p4) < 0.4)),
            op.NOT(op.rng_any(muons, lambda mu: op.deltaR(mu.p4, j.p4) < 0.4))))

        plots.append(Plot.make1D("nElectron", op.rng_len(electrons), noSel,
                                 EqB(10, 0., 10.), title="Number of electrons"))
        plots.append(Plot.make1D("nMuon", op.rng_len(muons), noSel,
                                 EqB(10, 0., 10.), title="Number of muons"))
        plots.append(Plot.make1D("nJet", op.rng_len(cleanedJets), noSel,
                                 EqB(10, 0., 10.), title="Number of jets"))

        hasOneJet = noSel.refine("hasOneJet", cut=(op.rng_len(cleanedJets) > 0))
        plots.append(Plot.make1D("jet1PT", cleanedJets[0].pt, hasOneJet,
                                 EqB(50, 0., 250.), title="Leading jet PT"))

        plots.append(Plot.make1D(
            "gen_nDaug", op.map(t.genpart, lambda gp: op.rng_len(gp.children)),
            noSel, EqB(100, 0., 100.), title="Number of children"))

        g_tquarks = op.select(t.genpart, lambda gp: op.abs(gp.pid) == 6)
        plots.append(Plot.make1D(
            "gen_nTop", op.rng_len(g_tquarks), noSel,
            EqB(100, 0., 100.), title="Number of top quark gen-particles"))
        g_dzero = op.select(t.genpart, lambda gp: op.abs(gp.pid) == 421)
        g_kshort = op.select(t.genpart, lambda gp: op.abs(gp.pid) == 310)
        plots += [
            Plot.make1D("gen_nD0S", op.rng_len(g_dzero), noSel,
                        EqB(20, 0., 20.), title="Number of D0 gen-particles"),
            Plot.make1D(
                "gen_D0ndaugs", op.map(g_dzero, lambda gp: op.rng_len(gp.children)),
                noSel, EqB(20, 0., 20.), title="D0 number of daughters"),
            Plot.make1D(
                "gen_D0ndesc", op.map(g_dzero, lambda gp: op.rng_count(gp.descendants)),
                noSel, EqB(20, 0., 20.), title="D0 number of descendants"),
            Plot.make2D(
                "gen_D0ndaugsndesc", (
                    op.map(g_dzero, lambda gp: op.rng_len(gp.children)),
                    op.map(g_dzero, lambda gp: op.rng_count(gp.descendants))
                ), noSel, (EqB(20, 0., 20.), EqB(20, 0., 20.)),
                title="D0 number of daughters vs descendants"),
            Plot.make1D(
                "gen_K0sndesc", op.map(g_kshort, lambda gp: op.rng_count(gp.descendants)),
                noSel, EqB(20, 0., 20.), title="K0_s number of descendants"),
        ]

        return plots



#!/usr/bin/env python3
import logging
import os
import os.path

logger = logging.getLogger(__name__)

PRINT_LEVELS = ["directory", "file", "histogram", "bin"]
LV_DIR = 0
LV_FILE = 1
LV_HISTOGRAM = 2
LV_BIN = 3


class LoggingDiffCollector:
    def __init__(self, level=0):
        self.level = level
        self.hasDiff = False

    def add(self, msg, level=0):
        self.hasDiff = True
        if level <= self.level:
            logger.error(msg)


def diffHistos(hA, hB, name=None, printLevel=0):
    diffs = LoggingDiffCollector(printLevel)
    nBinsOK, nBinsD = 0, 0
    assert hA.GetNcells() == hB.GetNcells()
    nBins = hA.GetNcells()
    for i in range(nBins):
        bcA = hA.GetBinContent(i)
        bcB = hB.GetBinContent(i)
        if bcA != bcB:
            diffs.add(f"Difference in bin {i:d} of histogram {name}: {bcA:e} (ref) vs {bcB:e} (test)", LV_BIN)
            nBinsD += 1
        else:
            nBinsOK += 1
    if nBinsD:
        diffs.add(f"{nBinsD:d} bins different for histogram {name} ({nBinsOK:d} equal)", LV_HISTOGRAM)
    return diffs.hasDiff


def diffFiles(fA, fB, name=None, skipSystematicVariations=False, printLevel=0):
    diffs = LoggingDiffCollector(printLevel)
    from cppyy import gbl
    histosA = {ky.GetName() for ky in fA.GetListOfKeys() if isinstance(ky.ReadObj(), gbl.TH1)}
    histosB = {ky.GetName() for ky in fB.GetListOfKeys() if isinstance(ky.ReadObj(), gbl.TH1)}
    if skipSystematicVariations:
        histosA = {nm for nm in histosA if "__" not in nm}
        histosB = {nm for nm in histosB if "__" not in nm}
    aNotB = histosA - histosB
    bNotA = histosB - histosA
    if aNotB:
        diffs.add(f"Histograms in reference but not in test for {name}: {', '.join(aNotB)}", LV_FILE)
    if bNotA:
        diffs.add(f"Histograms in test but not in reference for {name}: {', '.join(bNotA)}", LV_FILE)
    cmHistoNames = list(histosA.intersection(histosB))
    nHistosOK, nHistosD = 0, 0
    for hName in cmHistoNames:
        hA = fA.Get(hName)
        hB = fB.Get(hName)
        if diffHistos(hA, hB, name=f"{hName} in file {name}", printLevel=printLevel):
            nHistosD += 1
        else:
            nHistosOK += 1
    if nHistosD:
        diffs.add(f"{nHistosD:d} histograms with differences in file {name} ({nHistosOK:d} equal)", LV_FILE)
    return diffs.hasDiff


def collectFiles(dirX):
    fileNames = []
    for root, _dirs, files in os.walk(dirX):
        for f in files:
            if not f.startswith("__"):
                fileNames.append(os.path.relpath(os.path.join(root, f), dirX))
    return fileNames


def diffResultsDirs(test_dir, ref_dir, skipSystematicVariations=False, printLevel=0):
    diffs = LoggingDiffCollector(printLevel)
    fileNamesA = set(collectFiles(ref_dir))
    fileNamesB = set(collectFiles(test_dir))
    aNotB = fileNamesA - fileNamesB
    bNotA = fileNamesB - fileNamesA
    if aNotB:
        diffs.add(f"Files in reference but not in test: {', '.join(aNotB)}", LV_DIR)
    if bNotA:
        diffs.add(f"Files in test but not in reference: {', '.join(bNotA)}", LV_DIR)
    cmFileNames = list(fileNamesA.intersection(fileNamesB))
    nFilesOK, nFilesD = 0, 0
    from cppyy import gbl
    for fn in cmFileNames:
        fA = gbl.TFile.Open(os.path.join(ref_dir, fn))
        fB = gbl.TFile.Open(os.path.join(test_dir, fn))
        if diffFiles(fA, fB, name=fn, printLevel=printLevel,
                     skipSystematicVariations=skipSystematicVariations):
            nFilesD += 1
        else:
            nFilesOK += 1
    if nFilesD:
        diffs.add(f"{nFilesD:d} files with differences ({nFilesOK:d} equal)", LV_DIR)
    return diffs.hasDiff


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(
        description="Compare histograms in all ROOT files between two directories")
    parser.add_argument("test_dir", help="Test directory with histograms")
    parser.add_argument("reference_dir", help="Reference directory with histograms")
    parser.add_argument("--print-level", default="bin", choices=PRINT_LEVELS,
                        help="Printout detail level")
    parser.add_argument("--skip-systematic-variations", action="store_true",
                        help="Skip systematic variation histograms")
    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)
    if diffResultsDirs(
            args.test_dir, args.reference_dir,
            skipSystematicVariations=args.skip_systematic_variations,
            printLevel=PRINT_LEVELS.index(args.print_level)):
        import sys
        sys.exit(1)



"""
ROOT::RDataFrame backend classes
"""
import logging
import os.path
import shutil
import subprocess
import tempfile
from collections import defaultdict
from functools import partial
from itertools import chain
from timeit import default_timer as timer

import numpy as np
import pkg_resources

import bamboo.plots

from . import root
from . import treeoperations as top
from .plots import FactoryBackend, Selection, SelectionWithSub, Skim

# copied from https://docs.python.org/3/howto/logging-cookbook.html#use-of-alternative-formatting-styles


class Message:
    def __init__(self, fmt, args):
        self.fmt = fmt
        self.args = args

    def __str__(self):
        return self.fmt.format(*self.args)


class StyleAdapter(logging.LoggerAdapter):
    def __init__(self, logger, extra=None):
        super().__init__(logger, extra or {})

    def log(self, level, msg, *args, **kwargs):
        if self.isEnabledFor(level):
            msg, kwargs = self.process(msg, kwargs)
            self.logger._log(level, Message(msg, args), (), **kwargs)


logger = StyleAdapter(logging.getLogger(__name__))

_RDFNodeStats = defaultdict(int)
_RDFHistoNDStats = defaultdict(int)
_RDFHistoND_methods = dict()
# _RDFHistoND_methods[kyTypes] = getattr(gbl.rdfhelpers.rdfhistofactory,
#                                        f"Histo{nVars:d}D")[tuple(templTypes)]


def _printStats():
    logger.info(
        "Number of uses per node type: {}".format(
            ", ".join(f"{num:d} {ndTp}" for ndTp, num in _RDFNodeStats.items())))
    if _RDFHistoNDStats:
        logger.debug(
            "HistoND calls per column type: {}".format(
                ", ".join("{1:d} {0} -> {2};{3}".format(
                    kyTypes[0], num, ",".join(kyTypes[1]),
                    (kyTypes[2] if len(kyTypes) > 2 else ""))
                    for kyTypes, num in _RDFHistoNDStats.items())))
    if _RDFHistoND_methods:
        logger.debug(
            "HistoND helper instantiations: {}".format(
                ", ".join("{0}::Histo{2:d}D<{1}>".format(
                    kyTypes[0],
                    ",".join(list(kyTypes[1]) + ([kyTypes[2]] if len(kyTypes) > 2 else [])),
                    len(kyTypes[1]))
                    for kyTypes in _RDFHistoND_methods.keys())))


def _resetStats():
    global _RDFNodeStats
    global _RDFHistoNDStats
    global _RDFHistoND_methods
    _RDFNodeStats = defaultdict(int)
    _RDFHistoNDStats = defaultdict(int)
    _RDFHistoND_methods = dict()


class ExprWithName:
    """ Small helper to keep track of already-defined columns (with their name) """
    __slots__ = ("name", "op")

    def __init__(self, name, op):
        self.name = name
        self.op = op


class FilterWithDefines(top.CppStrRedir):
    """
    Branching point in the RDF graph: a Filter and the Define nodes attached to it

    Only when cuts are added, branching is done (may be multiple, if systematics-dependent).
    A proto-RDF-graph of FilterWithDefines nodes is built up by the backend as selections
    and plots are construction, but construction of the actual RDF nodes can be delayed
    (explicitly defined columns are also kept track of, while other 'shouldDefine' columns
    are still defined on-demand).
    """
    def __init__(self, parent, cut=None):
        self.explDefines = dict()
        self.var = None  # nodes for related systematic variations (if different by more than the weight)
        self.children = dict()
        if isinstance(parent, FilterWithDefines):
            self.parent = parent
            self.backend = parent.backend
            self.cut = cut
            self.digest = parent.digest  # updated if there are cuts (see below)
        elif isinstance(parent, DataframeBackend):
            self.parent = None
            self.backend = parent
            self.digest = top.op_digest("RDF::Filter")
        else:
            raise RuntimeError(
                "Can only define FilterWithDefines with a DataframeBackend "
                "or another FilterWithDefines")
        self._df = None
        self._definedColumns = dict()
        top.CppStrRedir.__init__(self)

    @property
    def df(self):
        if self._df is None:
            if self.parent is None:
                self._df = self.backend.rootDF
            else:
                self.parent.df  # construct parent RDF nodes first (if still needed)
                # trick: by doing this *before* constructing the nominal node, any definitions
                # end up in the node above, and are in principle available for other sub-selections too
                cutStr = self.cut.get_cppStr(defCache=self.parent)
                logger.debug(
                    f"FilterWithDefines with digest {top.hexStr_digest(self.digest)}: "
                    f"filtering with {cutStr}")
                self._df = self.backend.df_filter(self.parent.df, cutStr, self.cut, self.parent)
                # important - columns defined later with the parent will be invisible to the children
                self._definedColumns = dict(self.parent._definedColumns)
            for expr, nm in self.explDefines.items():
                if not self.getColName(expr):
                    self.define(expr, name=nm)
        return self._df

    @df.setter
    def df(self, arg):
        self._df = arg

    def childWithCuts(self, cuts):
        cutExpr = Selection._makeExprAnd(cuts)
        if cutExpr not in self.children:
            # branching the RDataFrame graph here (lazily)
            newNd = FilterWithDefines(self, cutExpr)
            newNd.digest = top.op_digest(self.digest, cutExpr)
            self.children[cutExpr] = newNd
            logger.debug(f"Created child FilterWithDefines  with digest {top.hexStr_digest(newNd.digest)}")
        else:
            logger.debug(f"Re-using child FilterWithDefines with digest {top.hexStr_digest(self.digest)}")
        return self.children[cutExpr]

    def getColName(self, op):
        if op in self._definedColumns:
            return self._definedColumns[op]
    stop = getColName  # column depedendencies

    def expl_define(self, expr, name=None):
        """ explicitly define column for expression """
        if self._df is not None and not self.getColName(expr):
            self.define(expr, name)
        elif expr not in self.explDefines:
            self.explDefines[expr] = name

    def shouldDefine(self, arg):
        return self.backend.shouldDefine(arg, defCache=self)

    def define(self, expr, name=None):
        """ Define as a new column (not checked, use getColName first) """
        if name is None:
            name = self.backend.getUColName()
        if expr in self._definedColumns:
            cppStr = self(expr)  # corner case: define with a new name
        else:
            cppStr = expr.get_cppStr(defCache=self)
        self.df = self.backend.df_define(self.df, name, cppStr, expr, self)
        self._definedColumns[expr] = name
        return name

    def symbol(self, decl, **kwargs):
        return self.backend.symbol(decl, **kwargs)

    def __call__(self, arg):
        """ Get the C++ string corresponding to an op """
        if not self._df:
            self.df  # need to be constructed here
        nm = self.getColName(arg)
        if not nm:
            if not self.shouldDefine(arg):
                try:
                    nm = arg.get_cppStr(defCache=self)
                except Exception as ex:
                    logger.exception("Could not get cpp string for {0!r}: {1!r}", arg, ex)
                    nm = "NONE"
            else:
                nm = self.define(arg)
        return nm

    def makeWeight(self, weights, wName=None, parentWeight=None):
        if parentWeight is not None:
            weights = [parentWeight] + list(weights)
        weight = Selection._makeExprProduct(weights)
        if not (isinstance(weight, top.GetColumn) or (isinstance(weight, top.ForwardingOp)
                                                      and isinstance(weight.wrapped, top.GetColumn))):
            self.expl_define(weight, name=wName)
        return weight


class SelectionHelper:
    # for lack of a better name... will keep the weight(s) for a selection and
    # the FilterWithDefines ref (can be the same for different Selections and SelectionHelpers)
    def __init__(self, fdnd):
        self.fdnd = fdnd  # FilterWithDefines-node reference
        self.weight = dict()


# NOTE these are global (for the current process&interpreter)
# otherwise they would be overwritten in sequential mode (this even allows reuse)
_gSymbols = dict()
_giFun = 0


def _normFunArgs(expr, args, argNames):
    newExpr = expr
    newArgs = args
    for i, argN in sorted(list(enumerate(argNames)), key=(lambda elm: len(elm[1])), reverse=True):
        newName = f"myArg{i:d}"
        if " " in argN:  # parameter-like, with a comment
            if sum(1 for ia in newArgs if argN in ia) != 1:  # should be in one and only one argument
                raise RuntimeError(f"{argN} is in more than one (or none) of {newArgs}")
            newArgs = [(ia.replace(argN, newName) if argN in ia else ia) for ia in newArgs]
        else:  # just a column name
            newArgs = [(" ".join((newName if tk == argN else tk) for tk in ia.split())
                        if argN in ia else ia) for ia in newArgs]
        newExpr = newExpr.replace(argN, newName)
    return newExpr, newArgs


def makeHistoName(plot, variation="nominal"):
    name = plot.name
    if variation != "nominal":
        name = "__".join((name, variation))
    return name


def _adjustBinnings(binnings):
    # high-dimensional histo models require either all or none of the binnings to be uniform
    if len(binnings) > 1:
        from .plots import EquidistantBinning, VariableBinning
        if any(isinstance(b, VariableBinning) for b in binnings):
            import numpy as np
            binnings = [VariableBinning(np.linspace(b.minimum, b.maximum, b.N + 1))
                        if isinstance(b, EquidistantBinning) else b for b in binnings]
    return binnings


class ProductHandle:
    NoResult = object()  # sentinel

    __slots__ = ("fdnd", "product")

    def __init__(self, filterWithDefNode):
        self.fdnd = filterWithDefNode
        self.product = None

    @property
    def digest(self):
        pass

    def makeProduct(self):
        pass

    def load(self, tFile):
        pass


class HistoND(ProductHandle):
    """ Lazy HistoND node: keep all information, but only build the RDF graph later """

    __slots__ = ("name", "fdnd", "allVars", "plot")

    def __init__(self, filterWithDefNode, name, axVars, plot, weight=None):
        super().__init__(filterWithDefNode)
        self.name = name
        self.allVars = list(axVars)
        if weight is not None:
            self.allVars.append(weight)
        self.plot = plot

    @property
    def digest(self):
        return top.op_digest(
            "RDF::Histo{nAx:d}D",
            self.fdnd.digest,
            *self.allVars,
            *(repr(bn) for bn in self.plot.binnings)
        )

    def _makeNamedVars(self):
        namedVars = []
        for i, var in enumerate(self.allVars):
            if isinstance(var, top.GetColumn):
                namedVars.append(ExprWithName(var.name, var))
            elif isinstance(var, top.ForwardingOp) and isinstance(var.wrapped, top.GetColumn):
                namedVars.append(ExprWithName(var.wrapped.name, var.wrapped))
            else:
                if self.fdnd.getColName(var):
                    varNm = self.fdnd.getColName(var)
                else:
                    varNm = f"v{i:d}_{self.name}"
                    self.fdnd.define(var, name=varNm)
                namedVars.append(ExprWithName(varNm, var))
        return namedVars

    def makeProduct(self):
        # define the necessary variables
        self.fdnd.df  # create Filter and (forced) Define nodes, if not done so yet
        namedVars = self._makeNamedVars()
        nAx = len(self.plot.variables)
        digest = self.digest
        logger.debug(
            "Adding histogram {0} with variables {1}{2} (digest: {3})",
            self.name, ", ".join(iv.name for iv in namedVars[:nAx]),
            (f" and weight {namedVars[-1].name}" if len(self.allVars) != nAx else ""),
            top.hexStr_digest(digest)
        )
        # reduce JITting inside RDF
        axTypes = [self.fdnd.df.GetColumnType(iv.name) for iv in namedVars]
        # TODO check & warn with allVars[:,1]
        useExplicit = True
        from .root import gbl
        for iv in namedVars:
            try:
                tp = getattr(gbl, iv.op.typeName)
                if hasattr(tp, "value_type"):
                    useExplicit = False
            except AttributeError:
                pass
        if useExplicit and nAx < 3:  # only have templates for those
            ndCppName = self.fdnd.df.__class__.__cpp_name__
            if nAx != len(axTypes):
                kyTypes = (ndCppName, tuple(axTypes[:-1]), axTypes[-1])
            else:
                kyTypes = (ndCppName, tuple(axTypes))
            _RDFHistoNDStats[kyTypes] += 1
            if kyTypes not in _RDFHistoND_methods:
                templTypes = [ndCppName] + axTypes
                logger.debug(f"Declaring Histo{nAx:d}D helper for types {templTypes}")
                _RDFHistoND_methods[kyTypes] = getattr(gbl.rdfhelpers.rdfhistofactory,
                                                       f"Histo{nAx:d}D")[tuple(templTypes)]
            plotFun = partial(_RDFHistoND_methods[kyTypes], self.fdnd.df)
        else:
            logger.debug(f"Using Histo{nAx:d}D with type inference")
            plotFun = getattr(self.fdnd.df, f"Histo{nAx:d}D")
        _RDFNodeStats[f"Histo{nAx:d}D"] += 1
        plotModel = self.fdnd.backend.__class__.makePlotModel(self.plot, self.name)
        self.product = plotFun(plotModel, *(iv.name for iv in namedVars))

    def load(self, tFile):
        self.product = tFile.Get(self.name)


class SkimTreeHandle(ProductHandle):

    __slots__ = ("colNToKeep", "skim", "tmpDir", "outFile")

    def __init__(self, filterWithDefNode, colNToKeep, skim):
        super().__init__(filterWithDefNode)
        self.colNToKeep = colNToKeep
        self.skim = skim
        self.tmpDir = tempfile.TemporaryDirectory()
        self.outFile = os.path.join(
            self.tmpDir.name,
            (f"skim_{self.skim.name}.root" if self.skim.key == self.skim.name
                else f"skim_{'_'.join(self.key)}.root"))

    @property
    def digest(self):
        return top.op_digest(
            "RDF::Snapshot",
            self.fdnd.digest,
            *chain(self.colNToKeep, chain.from_iterable(
                [k, top.adaptArg(v)] for k, v in self.skim.definedBranches.items()))
        )

    class Product:
        def __init__(self, parent, result=None, obj=None):
            self.parent = parent
            self.result = result
            self.obj = obj

        def Write(self):
            if self.GetEntries() != 0:
                newTree = self.obj.CloneTree(-1, "fast")
                newTree.Write()

        def GetEntries(self):
            if self.result and not self.obj:
                self._load(fromResult=True)
            return self.obj.GetEntries()

        def _load(self, fromResult=False):
            from .root import gbl
            pwd = gbl.gDirectory.GetPath()
            if fromResult:
                self.result.GetValue()  # trigger event loop if needed
            self._f = gbl.TFile.Open(self.parent.outFile)
            self.obj = self._f.Get(self.parent.skim.treeName)
            gbl.gDirectory.cd(pwd)  # move back

    def _getColNamesToKeep(self):
        self.fdnd.df  # create Filter and (forced) Define nodes, if not done so yet
        defcolN = list(self.fdnd._definedColumns.values())
        if hasattr(self.fdnd.df, "GetDefinedColumnNames"):  # only for non-compiled
            s_df = {str(nm) for nm in self.fdnd.df.GetDefinedColumnNames()
                    if str(nm) != "_zero_for_stats"}
            dfnotba = s_df - set(defcolN)
            if dfnotba:
                logger.error(f"In dataframe but not bamboo: {', '.join(dfnotba)}")
                raise RuntimeError("Mismatch in defined columns, see above")

        for cn in self.colNToKeep:
            if cn in defcolN:
                raise RuntimeError(f"Requested column '{cn}' from input is a defined column")
        colNToKeep = list(self.colNToKeep)

        for dN, dExpr in self.skim.definedBranches.items():
            if dN not in colNToKeep and dN not in defcolN:
                self.fdnd.define(top.adaptArg(dExpr), name=dN)
                colNToKeep.append(dN)
        return colNToKeep

    def makeProduct(self):
        from .root import gbl
        colNToKeep_v = gbl.std.vector["std::string"]()
        for cn in self._getColNamesToKeep():
            colNToKeep_v.push_back(cn)

        df = self.fdnd.df  # with the columns defined
        if self.skim.maxSelected > 0:
            df = df.Range(self.skim.maxSelected)
        snapOpts = gbl.ROOT.RDF.RSnapshotOptions()
        snapOpts.fLazy = True
        _RDFNodeStats["Snapshot"] += 1
        self.product = SkimTreeHandle.Product(
            self, result=df.Snapshot(self.skim.treeName, self.outFile, colNToKeep_v, snapOpts))

    def load(self, _tFile):
        self.product = SkimTreeHandle.Product(self)
        self.product._load(fromResult=False)


class CFRCounter(bamboo.plots.Product):
    def __init__(self, name, title=None):
        super().__init__(name)
        self.variables = (None,)
        self.binnings = (bamboo.plots.EquidistantBinning(1, 0., 1.),)
        self.title = title if title else name


class DataframeBackend(FactoryBackend):
    def __init__(self, tree, nThreads=None):
        from .root import gbl
        if nThreads and ((not gbl.ROOT.IsImplicitMTEnabled())
                         or nThreads != gbl.ROOT.GetThreadPoolSize()):
            logger.info(f"Enabling implicit MT for {nThreads:d} threads")
            gbl.ROOT.EnableImplicitMT(nThreads)
        self._tree = tree  # avoid going out of scope
        self.rootDF = gbl.ROOT.RDataFrame(tree).Define("_zero_for_stats", "0")
        self.selDFs = dict()       # (selection name, variation) -> SelWithDefines
        self.products = dict()     # product name -> list of product handles
        self.allSysts = set()      # all systematic uncertainties and variations impacting any plot
        self._cfrMemo = defaultdict(dict)
        super().__init__()
        self._iCol = 0

    def _getUSymbName(self):
        global _giFun
        _giFun += 1
        return f"myFun{_giFun:d}"

    def getUColName(self):
        self._iCol += 1
        return f"myCol{self._iCol:06d}"

    def shouldDefine(self, op, defCache=None):
        return op.canDefine and (isinstance(op, top.RangeOp) or isinstance(op, top.DefineOnFirstUse))

    def define(self, op, selection):  # for forceDefine
        self.selDFs[selection.name].fdnd.expl_define(op)

    def symbol(self, decl, resultType=None, args=None, nameHint=None):
        isFunction = (resultType and args)
        if isFunction:  # then it needs to be wrapped in a function
            decl = "{result} <<name>>({args})\n{{\n  return {0};\n}};\n".format(
                decl, result=resultType, args=args)
        global _gSymbols
        if decl in _gSymbols:
            return _gSymbols[decl]
        else:
            if nameHint and nameHint not in _gSymbols.values():
                name = nameHint
            else:
                name = self._getUSymbName()
            _gSymbols[decl] = name
            self.int_declare(decl.replace("<<name>>", name), isFunction=isFunction)
            return name

    def makeHelperFunction(self, cppStr, expr, defCache, asLambda=False):
        depList = top._collectDeps(expr, defCache=defCache)
        _, paramDecl, paramNames, paramCall = top._convertFunArgs(depList, defCache=defCache)
        expr_n, paramDecl_n = _normFunArgs(cppStr, paramDecl, paramNames)
        if defCache.getColName(expr) or any(isinstance(expr, tp)
                                            for tp in (top.GetColumn, top.GetArrayColumn)):
            _, paramDecl, paramNames, paramCall = top._convertFunArgs([expr], defCache=defCache)
            expr_n, paramDecl_n = _normFunArgs(cppStr, paramDecl, paramNames)
            funDecl = "[] ( {0} ) {{ return {1}; }}".format(
                ", ".join(paramDecl_n), expr_n)
            return funDecl, paramCall
        elif not asLambda:
            funName = defCache.symbol(expr_n, resultType=expr.typeName, args=", ".join(paramDecl_n))
            return funName, paramCall
        else:
            funDecl = "[] ( {0} ) {{ return {1}; }}".format(
                ", ".join(paramDecl_n), expr_n)
            return funDecl, paramCall

    # wrappers for gInterpreter::Declare, RDF::Define and RDF::Filter
    def int_declare(self, code, isFunction=False):
        logger.debug("Defining new symbol with interpreter: {0}", code)
        from .root import gbl
        gbl.gInterpreter.Declare(code)
        _RDFNodeStats["gInterpreter_Declare"] += 1

    def df_define(self, df, name, cppStr, expr, defCache):
        if isinstance(expr, top.RangeOp) and not defCache.getColName(expr):
            funName, param = self.makeHelperFunction(cppStr, expr, defCache)
            cppStr = f"{funName}({', '.join(param)})"
        logger.debug("Defining {0} as {1}", name, cppStr)
        _RDFNodeStats["Define"] += 1
        return df.Define(name, cppStr)

    def df_filter(self, df, cppStr, expr, defCache):
        if isinstance(expr, top.RangeOp) and not defCache.getColName(expr):
            funName, param = self.makeHelperFunction(cppStr, expr, defCache)
            cppStr = f"{funName}({', '.join(param)})"
        logger.debug("Filtering with {0}", cppStr)
        _RDFNodeStats["Filter"] += 1
        return df.Filter(cppStr)

    def df_origColNames(self):
        return [str(cN) for cN in self.rootDF.GetColumnNames() if str(cN) != "_zero_for_stats"]

    @classmethod
    def create(cls, decoTree, **kwargs):
        inst = cls(decoTree._tree, **kwargs)
        rootSel = Selection(inst, "none")
        return inst, rootSel

    def addSelection(self, sele):
        """ Define ROOT::RDataFrame objects needed for this selection """
        if sele.name in self.selDFs:
            raise RuntimeError(f"A Selection with name '{sele.name}' already exists")
        nomParentNd, parentHelper = None, None
        if sele.parent is None:  # root no-op selection
            assert not sele._cuts  # this one cannot have cuts
            logger.debug("Constructing FilterWithDefines for the root selection")
            nomNd = FilterWithDefines(self)
            nomNd.var = dict()  # will stay empty
            helper = SelectionHelper(nomNd)
        else:
            parentHelper = self.selDFs[sele.parent.name]
            nomParentNd = parentHelper.fdnd
            if sele._cuts:
                nomNd = nomParentNd.childWithCuts(sele._cuts)
                nomNd.var = dict()  # only for nominal non-root
            else:
                nomNd = nomParentNd
            helper = SelectionHelper(nomNd)
        nomParWeight = None
        if parentHelper and parentHelper.weight:
            nomParWeight = parentHelper.weight["nominal"]
        if sele._weights:
            helper.weight["nominal"] = nomNd.makeWeight(
                sele._weights,
                wName=f"w_{sele.name}",
                parentWeight=nomParWeight
            )
        elif nomParWeight:
            helper.weight["nominal"] = nomParWeight
        self.selDFs[sele.name] = helper

        if sele.autoSyst:
            # construct variation nodes (if necessary)
            seleSysts = sele.systematics
            logger.debug("Adding systematic variations {0}", ", ".join(seleSysts))
            for varn in seleSysts:  # all (cut and weight) variations
                # _cSysts and _wSysts are { variation : { cut/weight : [ nodes to change ] } }
                # add cuts to the appropriate node, if affected by systematics (here or up)
                varParentNd = None  # set parent node if not the nominal one
                if nomParentNd and varn in nomParentNd.var:  # -> continue on branch
                    logger.debug("{0} systematic variation {1}: continue on branch", sele.name, varn)
                    varParentNd = nomParentNd.var[varn]
                elif varn in sele._cSysts:  # -> branch off now
                    logger.debug("{0} systematic variation {1}: create branch", sele.name, varn)
                    varParentNd = nomParentNd
                if not varParentNd:  # cuts unaffected (here and in parent), can stick with nominal
                    varNd = nomNd
                else:  # on branch, so add cuts (if any)
                    if len(sele._cuts) == 0:  # no cuts, reuse parent
                        varNd = varParentNd
                    else:
                        ctMod = []
                        for ct in sele._cuts:
                            if ct in sele._cSysts[varn]:
                                clNds = list()
                                newct = ct.clone(select=sele._cSysts[varn][ct].__contains__,
                                                 selClones=clNds)
                                assert len(clNds) >= len(sele._cSysts[varn][ct])  # should clone all
                                for nd in clNds:
                                    nd.changeVariation(varn)
                                ctMod.append(newct)
                            else:
                                ctMod.append(ct)
                        varNd = varParentNd.childWithCuts(ctMod)
                    nomNd.var[varn] = varNd
                # next: attach weights (modified if needed) to varNd
                parw = parentHelper.weight.get(varn, parentHelper.weight.get("nominal"))
                if not sele._weights:
                    logger.debug("{0} systematic variation {1}: reusing {2}", sele.name, varn,
                                 ("parent weight" if parw is not None else "none"))
                    helper.weight[varn] = parw
                else:
                    if (varn in sele._wSysts or varNd != nomNd
                            or (parentHelper and varn in parentHelper.weight)):
                        wfMod = []
                        for wf in sele._weights:
                            if wf in sele._wSysts[varn]:
                                clNds = list()
                                newf = wf.clone(select=sele._wSysts[varn][wf].__contains__,
                                                selClones=clNds)
                                assert len(clNds) >= len(sele._wSysts[varn][wf])  # should clone all
                                for nd in clNds:
                                    nd.changeVariation(varn)
                                wfMod.append(newf)
                            else:
                                wfMod.append(wf)
                        logger.debug(
                            "{0} systematic variation {1}: defining new weight based on {2}",
                            sele.name, varn, parw)
                        helper.weight[varn] = varNd.makeWeight(
                            wfMod,
                            wName=(f"w_{sele.name}__{varn}" if sele._weights else None),
                            parentWeight=parw)
                    else:  # varNd == nomNd, not branched, and parent does not have weight variation
                        logger.debug(
                            "{0} systematic variation {1}: reusing nominal {2}",
                            sele.name, varn,
                            (varNd.weight["nominal"][0] if varNd.weight["nominal"] is not None
                             else "none"))
                        helper.weight[varn] = helper.weight["nominal"]

    def addPlot(self, plot, autoSyst=True):
        """
        Define ROOT::RDataFrame objects needed for this plot (and keep track of the result pointer)
        """
        if plot.key in self.products:
            raise ValueError(f"A Plot with key '{plot.key}' has already been added")

        selHelper = self.selDFs[plot.selection.name]
        nomNd = selHelper.fdnd
        plotRes = []
        # Add nominal plot
        nomWeight = selHelper.weight.get("nominal")
        if plot._weights:
            logger.debug(
                "Plot {0}: defining new nominal weight based on {1}",
                plot.name, ("selection nominal" if nomWeight is not None else "none"))
            nomWeight = nomNd.makeWeight(
                plot._weights, parentWeight=nomWeight,
                wName=f"w_{plot.selection.name}_{plot.name}")
        plotRes.append(HistoND(nomNd, makeHistoName(plot), plot.variables, plot, weight=nomWeight))

        if plot.selection.autoSyst and autoSyst:
            # Same for all the systematics
            varSysts = top.collectSystVars(plot.variables)
            pwSysts = top.collectSystVars(plot._weights)
            # plotSysts, selSysts and allSysts are just {varName}, without the nodes to change
            plotSysts = set(varSysts)
            plotSysts.update(pwSysts)
            selSysts = plot.selection.systematics
            allSysts = set(selSysts)
            allSysts.update(plotSysts)
            self.allSysts.update(allSysts)
            for varn in allSysts:
                if varn in varSysts or varn in nomNd.var:
                    varNd = nomNd.var.get(varn, nomNd)
                    vVars = []
                    for xVar in plot.variables:
                        if xVar in varSysts[varn]:
                            clNds = list()
                            newVar = xVar.clone(select=varSysts[varn][xVar].__contains__,
                                                selClones=clNds)
                            assert len(clNds) >= len(varSysts[varn][xVar])  # should clone all
                            for nd in clNds:
                                nd.changeVariation(varn)
                            vVars.append(newVar)
                        else:
                            vVars.append(xVar)
                else:  # can reuse variables, but need to take care of weight
                    varNd = nomNd
                    vVars = plot.variables
                weight = selHelper.weight[varn if varn in selSysts and varn in selHelper.weight
                                          else "nominal"]
                if plot._weights:
                    wfMod = []
                    if varn in pwSysts:
                        # plot weights have variation, vary
                        for wf in plot._weights:
                            if wf in pwSysts[varn]:
                                clNds = list()
                                newf = wf.clone(select=pwSysts[varn][wf].__contains__,
                                                selClones=clNds)
                                assert len(clNds) >= len(pwSysts[varn][wf])  # should clone all
                                for nd in clNds:
                                    nd.changeVariation(varn)
                                wfMod.append(newf)
                            else:
                                wfMod.append(wf)
                    elif varn in selSysts and varn in selHelper.weight:
                        # variation in selection weight
                        wfMod = plot._weights
                    if wfMod:  # need a weight for this variation
                        logger.debug(
                            "Plot {0} systematic variation {1}: defining new weight based on {2}",
                            plot.name, varn, ("parent weight" if weight is not None else "none"))
                        weight = varNd.makeWeight(
                            wfMod, parentWeight=weight,
                            wName=f"w_{plot.selection.name}_{plot.name}__{varn}")
                    else:  # no systematic that affects the weight
                        weight = nomWeight
                plotRes.append(HistoND(varNd, makeHistoName(plot, variation=varn), vVars,
                                       plot, weight=weight))

        self.products[plot.key] = plotRes

    @classmethod
    def makePlotModel(cls, plot, hName):
        from .root import gbl
        modCls = getattr(gbl.ROOT.RDF, f"TH{len(plot.binnings):d}DModel")
        return modCls(hName, plot.title, *chain.from_iterable(
            cls.makeBinArgs(binning) for binning in _adjustBinnings(plot.binnings)))

    @classmethod
    def makeBinArgs(cls, binning):
        from .plots import EquidistantBinning, VariableBinning
        if isinstance(binning, EquidistantBinning):
            return (binning.N, binning.mn, binning.mx)
        elif isinstance(binning, VariableBinning):
            return (binning.N, np.array(binning.binEdges, dtype=np.float64))
        else:
            raise ValueError(f"Binning of unsupported type: {binning!r}")

    def getProducts(self, key):
        return self.products.get(key, [])

    def getResults(self, plot, key=None):
        bareResults = []
        for res in self.getProducts(key if key is not None else plot.key):
            if isinstance(res, ProductHandle):
                prod = res.product
                if prod is not ProductHandle.NoResult:
                    bareResults.append(prod)
            else:  # TODO get rid of this check when everything follows the above
                logger.warning(
                    f"getProducts: Adding result {res!r} for "
                    f"{key if key is not None else plot.key} directly")
                bareResults.append(res)
        return plot.produceResults(bareResults, self, key=key)

    def addSkim(self, skim):
        if skim.key in self.products:
            raise ValueError(f"A Skim with key '{skim.key}' has already been added")

        colNToKeep = list()
        origcolN = self.df_origColNames()
        for toKeep in skim.originalBranches:
            if toKeep is Skim.KeepAll:  # keep all if not defined
                for cN in origcolN:
                    if cN not in colNToKeep:
                        colNToKeep.append(cN)
            elif isinstance(toKeep, Skim.KeepRegex):
                import re
                pat = re.compile(toKeep.pattern)
                for cN in origcolN:
                    if pat.match(cN) and cN not in colNToKeep:
                        colNToKeep.append(cN)
            elif isinstance(toKeep, str):
                if toKeep not in origcolN:
                    raise RuntimeError(f"Requested column '{toKeep}' from input not found")
                if toKeep not in colNToKeep:
                    colNToKeep.append(toKeep)
            else:
                raise RuntimeError(f"Cannot interpret {toKeep!r} as column list to keep")

        for dN in skim.definedBranches.keys():
            if dN in origcolN:
                logger.warning(
                    f"Requested to add column {dN} with expression, but a column with the same "
                    "name on the input tree exists. The latter will be copied instead")

        selND = self.selDFs[skim.selection.name].fdnd

        self.products[skim.key] = [SkimTreeHandle(selND, colNToKeep, skim)]

    def addCutFlowReport(self, report, selections, key=None, autoSyst=True):
        logger.debug(
            "Adding cutflow report {} for selection(s) {}".format(
                report.name, ", ".join(sele.name for sele in selections.values())))
        memo = self._cfrMemo[report.name]
        cfres, results = [], []
        for name, sele in selections.items():
            if sele.name in memo:
                cfr = memo[sele.name]
            else:
                cfrNom, cfrSys = self._makeCutFlowReport(sele, name, autoSyst=autoSyst,
                                                         prefix=report.name)
                cfr = report.__class__.Entry(name, cfrNom, cfrSys)
                memo[sele.name] = cfr
            cfres.append(cfr)
            results.append(cfr.nominal)
            results += list(cfr.systVars.values())
        if key is None:
            key = report.name
        report.cfres[key] += cfres
        self.products[key] = self.getProducts(key) + results

    def _makeCutFlowReport(self, selection, name, autoSyst=True, prefix=None):
        selH = self.selDFs[selection.name]
        selND = selH.fdnd
        nomWeight = selH.weight.get("nominal")
        nomName = f"{prefix}_{name}"
        zVar = [top.GetColumn("int", "_zero_for_stats")]
        cfrNom = HistoND(
            selND, nomName, zVar + ([nomWeight] if nomWeight is not None else []),
            CFRCounter(nomName, f"CutFlowReport {prefix} nominal counter for {selection.name}"))
        cfrSys = {}
        if autoSyst:
            for varNm in selection.systematics:
                if varNm in selND.var or selH.weight[varNm] != nomWeight:
                    cfrSys[varNm] = HistoND(
                        selND.var.get(varNm, selND),
                        f"{nomName}__{varNm}", zVar + [selH.weight[varNm]],
                        CFRCounter(
                            f"{nomName}__{varNm}",
                            f"CutFlowReport {prefix} {varNm} counter for {selection.name}"))
        return cfrNom, cfrSys

    def buildGraph(self, plotList):
        self._plotList = plotList  # store for runGraph
        # gather products
        products = defaultdict(list)
        for p in plotList:
            products[None] += self.getProducts(p.name)
            for suff in SelectionWithSub.getSubsForPlot(p, requireActive=True, silent=True):
                products[suff] += self.getProducts((p.name, suff))
        # Collect FilterWithDefines that need to be built, and products for each
        products_by_fdndId = dict()
        fdnds = dict()
        for sProducts in products.values():
            for prod in sProducts:
                # (later) figure out which are there, and which need to be produced
                assert prod.product is None  # need to be all
                fdndId = id(prod.fdnd)
                cId, cNd = fdndId, prod.fdnd
                while cId not in fdnds and isinstance(cNd, FilterWithDefines):
                    fdnds[cId] = cNd
                    products_by_fdndId[cId] = []
                    cNd = cNd.parent
                    cId = id(cNd)
                products_by_fdndId[fdndId].append(prod)

        # strategy from old lazy backend: start from top, build products before children
        def buildFilterWithDefines(iId, iNd):
            for prod in products_by_fdndId[iId]:
                prod.makeProduct()  # RDF graph construction
            del products_by_fdndId[iId]
            for cNd in fdnds[iId].children.values():
                cId = id(cNd)
                if cId in products_by_fdndId:
                    buildFilterWithDefines(cId, cNd)
            del fdnds[iId]

        buildFilterWithDefines(*next((k, v) for k, v in fdnds.items() if v.parent is None))
        assert not products_by_fdndId

        _printStats()
        _resetStats()
        # NOTE could return results

    def runGraph(self):
        from .root import gbl
        triggered = False
        for p in self._plotList:
            for h in self.getResults(p):
                if not triggered and not isinstance(h, gbl.TH1):
                    h.GetEntries()  # trigger event loop
                    triggered = True
            for suff in SelectionWithSub.getSubsForPlot(p, requireActive=True, silent=False):
                for h in self.getResults(p, key=(p.name, suff)):
                    if not triggered and not isinstance(h, gbl.TH1):
                        h.GetEntries()  # trigger event loop
                        triggered = True

    def writeResults(self, plotList, outputFile, mergeCounters):
        numHistos = 0
        numSkims = 0
        numDDHistos = 0
        from .root import gbl
        outF = gbl.TFile.Open(outputFile, "RECREATE")
        outF.cd()
        for p in plotList:
            for h in self.getResults(p):
                h.Write()
                if isinstance(h, SkimTreeHandle.Product):
                    numSkims += 1
                else:
                    numHistos += 1
        mergeCounters(outF)
        outF.Close()
        subFiles = dict()
        for p in plotList:
            for suff in SelectionWithSub.getSubsForPlot(p, requireActive=True, silent=True):
                if suff not in subFiles:
                    subFiles[suff] = gbl.TFile.Open(
                        "{1}{0}{2}".format(suff, *os.path.splitext(outputFile)),
                        "RECREATE")
                subFiles[suff].cd()
                for h in self.getResults(p, key=(p.name, suff)):
                    h.Write()
                    numDDHistos += 1
        for subF in subFiles.values():
            mergeCounters(subF)
            subF.Close()
        return {
            "nHistos": numHistos,
            "nSkims": numSkims,
            "nDDHistos": numDDHistos
        }

    def addDependency(self, **kwargs):
        from .root import loadDependency
        for ky in ("package", "cmakeArgs"):  # only for fully compiled
            if ky in kwargs:
                kwargs.pop(ky)
        return loadDependency(**kwargs)


class DebugDataframeBackend(DataframeBackend):
    """ Build the full RDF graph immediately, as plots and selections are being defined """

    def addSelection(self, selection):
        super().addSelection(selection)
        self.selDFs[selection.name].fdnd.df  # construct RDF graph until Filter node(s)

    def addPlot(self, plot, autoSyst=True):
        super().addPlot(plot, autoSyst=autoSyst)
        for p in self.getProducts(plot.key):
            p.makeProduct()

    def addSkim(self, skim):
        super().addSkim(skim)
        for p in self.getProducts(skim.key):
            p.makeProduct()

    def addCutFlowReport(self, report, selections, key=None, autoSyst=True):
        super().addCutFlowReport(report, selections, key=key, autoSyst=autoSyst)
        for prod in self.getProducts(key if key is not None else report.name):
            if prod.product is None:
                prod.makeProduct()

    def buildGraph(self, plotList):
        self._plotList = plotList  # store for runGraph
        _printStats()
        _resetStats()


class FullyCompiledBackend(DataframeBackend):
    """ Another experiment: compile a standalone executable to avoid any JITting in RDF """
    skeleton_cmake = pkg_resources.resource_filename("bamboo", os.path.join("data", "CMakeLists.txt"))
    skeleton_cpp = pkg_resources.resource_filename("bamboo", os.path.join("data", "main.cc"))

    def __del__(self):
        if (self.cleanupTmp and self.tmpDir and os.path.isdir(self.tmpDir)
                and self.tmpDir.startswith(tempfile.gettempdir())):
            logger.debug(f"Cleaning up temporary directory {self.tmpDir}")
            shutil.rmtree(self.tmpDir)
        else:
            logger.info(f"Leaving all generated code and cmake outputs in {self.tmpDir}")

    _genCode = {
        # cmake
        "findpackages": [],
        "bamboocomponents": list(root._defaultBambooLibs),
        "libdependencies": list(root._defaultDependencies) + list(root._defaultBambooLibs),
        "includedirs": [],
        "linkdirs": [],
        # cc
        "includes": [f'#include "hdr"' for hdr in root._defaultHeaders],
        "helpermethods": [],
        "initialisation": [],
    }

    gen_entriesFmt = {
        "includes": (list(root._defaultHeaders), lambda hdrs: [f'#include "{hdr}"' for hdr in hdrs]),
        "includedirs": ([], lambda dirs: [
            "target_include_directories(bambooExecutor PRIVATE {})".format(" ".join(
                f"$<BUILD_INTERFACE:{idr}>" for idr in dirs))]),
        "linkdirs": ([], lambda dirs: [
            "target_link_directories(bambooExecutor PUBLIC {})".format(" ".join(dirs))])
    }

    @staticmethod
    def _addGenEntries(key, entries):
        if key in FullyCompiledBackend.gen_entriesFmt:
            gEnt, fmt = FullyCompiledBackend.gen_entriesFmt[key]
            gEnt.extend(entries)
            FullyCompiledBackend._genCode[key] = fmt(gEnt)
        else:
            FullyCompiledBackend._genCode[key] += entries

    def __init__(self, tree, nThreads=None):
        super().__init__(tree)
        self.tTree = tree
        self.rootDF = "df0"
        self._iDF = 0
        self.nThreads = (nThreads if nThreads else 1)
        self.cmakeConfigOptions = ["-DCMAKE_BUILD_TYPE=Release"]
        self.cmakeBuildOptions = []
        self.cleanupTmp = True
        self.tmpDir = tempfile.mkdtemp()
        logger.debug(f"Temporary directory for compilation: {self.tmpDir}")
        self.tmpOutF = None
        self.resultDefinitions = ['auto df0 = df.Define("_zero_for_stats", []() { return 0; }, {});']
        self.inputFileLists = None

    @property
    def genCode(self):
        gc = dict(FullyCompiledBackend._genCode)
        gc["resultdefinitions"] = self.resultDefinitions
        return gc

    def getUDFName(self):
        self._iDF += 1
        return f"df{self._iDF:d}"

    def shouldDefine(self, op, defCache=None):
        return (op.canDefine and (isinstance(op, top.RangeOp)
                                  or isinstance(op, top.DefineOnFirstUse)
                                  or isinstance(op, top.Parameter)))

    def int_declare(self, code, isFunction=False):
        if not isFunction:  # no need to have helper functions dynamically
            super().int_declare(code, isFunction=isFunction)
        FullyCompiledBackend._genCode["helpermethods"].append(code)

    def df_define(self, df, name, cppStr, expr, defCache):
        funNameOrDef, param = self.makeHelperFunction(
            cppStr, expr, defCache, asLambda=(not isinstance(expr, top.RangeOp)))
        newdf = self.getUDFName()
        self.resultDefinitions.append(
            f'auto {newdf} = {df}.Define("{name}", {funNameOrDef}, '
            + "{{ {0} }});".format(", ".join(f'"{p}"' for p in param)))
        _RDFNodeStats["Define"] += 1
        return newdf

    def df_filter(self, df, cppStr, expr, defCache):
        funNameOrDef, param = self.makeHelperFunction(
            cppStr, expr, defCache,
            asLambda=((not isinstance(expr, top.RangeOp)) or defCache.getColName(expr)))
        newdf = self.getUDFName()
        self.resultDefinitions.append(
            f"auto {newdf} = {df}.Filter({funNameOrDef},"
            + "{{ {0} }});".format(", ".join(f'"{p}"' for p in param)))
        _RDFNodeStats["Filter"] += 1
        return newdf

    def df_origColNames(self):
        from .root import gbl
        return list(str(cN) for cN in gbl.ROOT.RDataFrame(self.tTree).GetColumnNames())

    @classmethod
    def makePlotModel(cls, plot, hName):
        binArgs = ", ".join(cls.makeBinArgs(bn) for bn in _adjustBinnings(plot.binnings))
        return f'ROOT::RDF::TH{len(plot.binnings):d}DModel("{hName}", "{plot.title}", {binArgs})'

    @classmethod
    def makeBinArgs(cls, binning):
        from .plots import EquidistantBinning, VariableBinning
        if isinstance(binning, EquidistantBinning):
            return f"{binning.N:d}, {binning.mn:f}, {binning.mx:f}"
        elif isinstance(binning, VariableBinning):
            return "{0:d}, {{ {1} }}".format(
                len(binning.N), ", ".join(str(be) for be in binning.binEdges))
        else:
            raise ValueError(f"Binning of unsupported type: {binning!r}")

    def _writeWithSubstitution(self, skeletonFile, placeholder):
        with open(skeletonFile) as skF:
            srcLines = []
            for skLn in skF:
                if skLn.lstrip().startswith(placeholder):
                    ind, ky = skLn.split(placeholder)
                    srcLines += [f"{ind}{ln}" for ln in self.genCode[ky.strip()]]
                else:
                    srcLines.append(skLn.rstrip("\n"))
        with open(os.path.join(self.tmpDir, os.path.basename(skeletonFile)), "w") as srcF:
            fullSrc = "\n".join(srcLines)
            logger.debug("Full source: \n{0}", fullSrc)
            srcF.write(fullSrc)

    def buildGraph(self, plotList):
        self._plotList = plotList  # store for runGraph
        # gather products
        products = defaultdict(list)
        for p in plotList:
            products[None] += self.getProducts(p.name)
            for suff in SelectionWithSub.getSubsForPlot(p, requireActive=True):
                products[suff] += self.getProducts((p.name, suff))
        # figure out which are there, and which need to be produced
        nToProduce, nFromCache = 0, 0
        for sProducts in products.values():
            for prod in sProducts:
                assert prod.product is None
                prod.fdnd.df  # create Filter and (forced) Define nodes, if not done so yet
                if isinstance(prod, HistoND):
                    namedVars = prod._makeNamedVars()
                    self.resultDefinitions.append(
                        f"results.add({prod.fdnd.df}.Histo{len(prod.plot.variables):d}D"
                        + "<{}>({}, {}));".format(
                            ", ".join(v.op.typeName for v in namedVars),
                            self.__class__.makePlotModel(prod.plot, prod.name),
                            ", ".join(f'"{v.name}"' for v in namedVars)))
                    _RDFNodeStats[f"Histo{len(prod.plot.variables):d}D"] += 1
                elif isinstance(prod, SkimTreeHandle):
                    colNToKeep = prod._getColNamesToKeep()  # also trigger column defines
                    self.resultDefinitions.append(
                        "results.add({}{}".format(
                            prod.fdnd.df,
                            (f".Range({prod.skim.maxSelected})"
                             if prod.skim.maxSelected > 0 else ""))
                        + f'.Snapshot("{prod.skim.treeName}", "{prod.outFile}", {{ '
                        + ", ".join(f'"{cN}"' for cN in colNToKeep)
                        + " }, snapshotOptions));"
                    )
                    _RDFNodeStats["Snapshot"] += 1
                else:
                    raise RuntimeError(f"Unsupported product type: {prod!r}")
                nToProduce += 1
        logger.info(
            f"{nFromCache:d}/{nToProduce+nFromCache:d} histograms from cache, "
            f"{nToProduce:d}/{nToProduce+nFromCache:d} to produce")
        _printStats()
        _resetStats()
        # write code and compile
        self._writeWithSubstitution(FullyCompiledBackend.skeleton_cmake, "# BAMBOO_INSERT")
        self._writeWithSubstitution(FullyCompiledBackend.skeleton_cpp, "// BAMBOO_INSERT")
        logger.info("Compiling executable")
        start = timer()
        try:
            configCmd = ["cmake"] + self.cmakeConfigOptions + [
                f'-Dbamboo_DIR={pkg_resources.resource_filename("bamboo", "cmake")}',
                f'-DCMAKE_MODULE_PATH={pkg_resources.resource_filename("bamboo", "data")}',
                self.tmpDir]
            out = subprocess.check_output(configCmd, cwd=self.tmpDir, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as ex:
            self.cleanupTmp = False
            logger.error("Problem while running cmake, full output of '{0}':\n{1}",
                         " ".join(configCmd), ex.output.decode().strip())
            raise ex
        else:
            logger.debug("Full output of '{0}:\n{1}", " ".join(configCmd), out.decode().strip())
        try:
            buildCmd = ["cmake", "--build", self.tmpDir] + self.cmakeBuildOptions
            out = subprocess.check_output(buildCmd, cwd=self.tmpDir, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as ex:
            self.cleanupTmp = False
            logger.error("Problem while building, full output of '{0}':\n{1}",
                         " ".join(buildCmd), ex.output.decode().strip())
            raise ex
        else:
            logger.debug("Full output of '{0}':\n{1}", " ".join(buildCmd), out.decode().strip())
        end = timer()
        logger.info(f"Executable compilation finished in {end - start:.2f}s")

    def runGraph(self):
        tmpOutFileName = os.path.join(self.tmpDir, "out.root")
        inFileArr = self.tTree.GetListOfFiles()  # assumes TChain
        inFileNames = [inFileArr.At(i).GetTitle() for i in range(inFileArr.GetEntries())]
        threadsMsg = (f" ({self.nThreads:d} threads)"
                      if self.nThreads is not None and self.nThreads > 1
                      else "")
        logger.info(f"Processing {len(inFileNames):d} files with standalone executable{threadsMsg}")
        try:
            cmd = [
                os.path.join(self.tmpDir, "bambooExecutor"),
                f"--output={tmpOutFileName}",
                f"--tree={self.tTree.GetName()}",
            ] + ([f"--threads={self.nThreads:d}"]
                 if self.nThreads is not None and self.nThreads > 1 else []
                 ) + ([f"--input-files={inFN}" for inFN in self.inputFileLists]
                      if self.inputFileLists is not None else
                      [f"--input={inFN}" for inFN in inFileNames])
            logger.info(f"Full command: {' '.join(cmd)}")
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as ex:
            self.cleanupTmp = False
            logger.error("Problem while running standalone executable, full output:\n{0}",
                         ex.output.decode())
            raise ex
        # then load them in the results dictionary to make the res work
        from .root import gbl, returnToDirectory
        with returnToDirectory():
            # keep a ref to the file because it owns the histograms
            # (and use to detect this part has run)
            self.tmpOutF = gbl.TFile.Open(tmpOutFileName)
            loadedRes = dict()
            for pNm, pResults in self.products.items():
                for res in pResults:
                    res.load(self.tmpOutF)
                loadedRes[pNm] = list(pResults)
            self.products = loadedRes

    def addDependency(self, bambooLib=None,
                      includePath=None, headers=None,
                      dynamicPath=None, libraries=None,
                      package=None, components=None,
                      cmakeArgs=None):
        # start with loading them as usual
        super().addDependency(
            bambooLib=bambooLib, includePath=includePath, headers=headers,
            dynamicPath=dynamicPath, libraries=libraries)

        def asList(arg):
            if isinstance(arg, str):
                return [arg]
            else:
                return list(arg)

        if includePath:
            FullyCompiledBackend._addGenEntries("includedirs", asList(includePath))
        if headers:
            FullyCompiledBackend._addGenEntries("includes", asList(headers))
        if dynamicPath:
            FullyCompiledBackend._addGenEntries("linkdirs", asList(dynamicPath))
        if libraries:
            FullyCompiledBackend._genCode["libdependencies"] += asList(libraries)
        if bambooLib:
            FullyCompiledBackend._genCode["bamboocomponents"] += asList(bambooLib)
            FullyCompiledBackend._genCode["libdependencies"] += asList(bambooLib)
        if package:
            FullyCompiledBackend._genCode["findpackages"].append("find_package({} REQUIRED{})".format(
                package, (
                    "" if not components else " COMPONENTS {}".format(", ".join(asList(components))))))
        if cmakeArgs:
            if isinstance(cmakeArgs, str):
                self.cmakeConfigOptions.append(cmakeArgs)
            else:
                self.cmakeConfigOptions += list(cmakeArgs)


def getBackend(decoTree, name=None, **kwargs):
    if name == "debug":
        cls = DebugDataframeBackend
    elif name == "compiled":
        cls = FullyCompiledBackend
    else:
        cls = DataframeBackend
    return cls.create(decoTree, **kwargs)



import os.path

import pkg_resources
import pytest

from helpers import isclose_float

pytest.importorskip("torch")
pytestmark = pytest.mark.skipif(
    not pkg_resources.resource_exists("bamboo", "lib/libBambooTorch.so"),
    reason="No libBambooTorch.so")

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def libtorch_nn1():
    from bamboo.root import gbl, loadLibTorch
    loadLibTorch()
    yield gbl.bamboo.TorchEvaluator(os.path.join(testData, "nn1.pt"))


def test_libtorch_load(libtorch_nn1):
    assert libtorch_nn1


def test_libtorch_evalZero(libtorch_nn1):
    from bamboo.root import gbl
    v_in = gbl.std.vector["float"]()
    for _i in range(5):
        v_in.push_back(.7)
    v_out = libtorch_nn1.evaluate(v_in)
    assert (v_out.size() == 2) and isclose_float(sum(v_out), 1.)


def test_libtorch_evalRandom(libtorch_nn1):
    from bamboo.root import gbl
    import random
    v_in = gbl.std.vector["float"]()
    for _i in range(5):
        v_in.push_back(random.uniform(0., 1.))
    v_out = libtorch_nn1.evaluate(v_in)
    assert (v_out.size() == 2) and isclose_float(sum(v_out), 1.)



import json
import logging
import os.path
import requests
import twine.package

logger = logging.getLogger(__name__)


def upload_zenodo(distribution_file, endpoint, access_token,
                  metadata=None, update=None, publish=False):
    pkg = twine.package.PackageFile.from_filename(distribution_file,
                                                  comment=None)
    pkg_meta = pkg.metadata_dictionary()
    combined_meta = {
        "upload_type": "software",
        "version": pkg_meta["version"],
        "license": pkg_meta["license"],
    }
    if metadata:
        if "description" not in metadata and pkg_meta["description"]:
            resp = requests.post(
                "https://api.github.com/markdown",
                params={"accept": "application/vnd.github.v3+json"},
                data=json.dumps({"text": pkg_meta["description"]})
            )
            if resp.ok:
                metadata["description"] = resp.content.decode()
        combined_meta.update(metadata)

    from pprint import pformat
    logger.debug(f"Metadata: {pformat(combined_meta)}")

    params = {"access_token": access_token}
    with requests.Session() as session:
        session.hooks = {
            "response": lambda r, *args, **kwargs: r.raise_for_status()
        }
        if not update:
            r = session.post(f"{endpoint}/deposit/depositions",
                             params=params, json={},
                             headers={"Content-Type": "application/json"})
            logger.debug(f"Deposition created response: {pformat(r.json())}")
            deposition = f"{endpoint}/deposit/depositions/{r.json()['id']}"
        else:
            r = session.post(f"{endpoint}/deposit/depositions/{update}"
                             "/actions/newversion", params=params)
            logger.debug(f"Deposition opened for version add: {pformat(r.json())}")
            deposition = r.json()["links"]["latest_draft"]
            r = session.get(deposition, params=params)
        bucket_link = r.json()["links"]["bucket"]
        with open(distribution_file, "rb") as fp:
            r = requests.put(
                "/".join((bucket_link, os.path.basename(distribution_file))),
                data=fp, params=params)
            logger.debug(f"File upload response: {pformat(r.json())}")

        r = requests.put(deposition, params=params,
                         data=json.dumps({"metadata": combined_meta}),
                         headers={"Content-Type": "application/json"})
        logger.debug(f"Metadata edit response: {pformat(r.json())}")

        logger.info("The draft deposit can be further edited or published at "
                    + r.json()["links"]["latest_draft_html"])
        if publish:
            r = requests.post(r.json()["links"]["publish"], params=params)
            logger.debug(f"Publication response: {pformat(r.json())}")


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description="Upload release to zenodo")
    parser.add_argument("file", help="Distribution file")
    parser.add_argument("--zenodo-url", default="https://zenodo.org/api",
                        help=("Zenodo API endpoint (typically "
                              "https://zenodo.org/api or "
                              "https://sandbox.zenodo.org/api)"))
    parser.add_argument("--access-token", required=True)
    parser.add_argument("--update",
                        help="ID of the deposition to add this version to")
    parser.add_argument("--metadata",
                        help="File with additional metadata for the upload")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Verbose printout")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="Reduced printout")
    parser.add_argument("--publish", action="store_true",
                        help="Publish the (updated) deposition")
    args = parser.parse_args()
    logging.basicConfig(level=(
        logging.DEBUG if args.verbose else (
            logging.WARNING if args.quiet else logging.INFO
        )
    ))
    metadata = {}
    if args.metadata:
        with open(args.metadata) as mF:
            metadata = json.load(mF)
    upload_zenodo(args.file, args.zenodo_url, args.access_token,
                  metadata=metadata, update=args.update, publish=args.publish)



import yaml

from cppyy import gbl
from dataframe.dataframebackend import DataframeBackend
from dataframe.treedecorators import decorateTTW

with open("ttW.yaml") as df:
    ttWdesc = yaml.load(df)

f = gbl.TFile.Open("ttW_sample.root")
t_ = f.Get("t")
t = decorateTTW(t_, description=ttWdesc)
be, noSel = DataframeBackend.create(t)



"""
HTCondor tools (based on cp3-llbb/CommonTools condorSubmitter)
"""
__all__ = ("CommandListJob", "jobsFromTasks", "makeTasksMonitor", "findOutputsForCommands")

import logging
import os
import os.path
import subprocess
from itertools import chain, count

from .batch import CommandListJob as CommandListJobBase

logger = logging.getLogger(__name__)


def makeExecutable(path):
    """ Set file permissions to executable """
    import stat
    if os.path.exists(path) and os.path.isfile(path):
        perm = os.stat(path)
        os.chmod(path, perm.st_mode | stat.S_IEXEC)


CondorJobStatus = [
    "UNEXPANDED",      # 0
    "IDLE",            # 1
    "RUNNING",         # 2
    "REMOVED",         # 3
    "COMPLETED",       # 4
    "HELD",            # 5
    "SUBMISSION_ERR",  # 6
    "FAILED"           # 7 NOTE: not in HTCondor
]


class CommandListJob(CommandListJobBase):
    """
    Helper class to create a condor master job from a list of commands (each becoming one subjob)

    Default work directory will be $(pwd)/condor_work, default output pattern is "*.root"
    """
    def __init__(self, commandList, workDir=None, cmdLines=None, envSetupLines=None, outputPatterns=None):
        self.envSetupLines = envSetupLines if envSetupLines is not None else []
        self.outputPatterns = outputPatterns if outputPatterns is not None else ["*.root"]

        super().__init__(commandList, workDir=workDir, workdir_default_pattern="condor_work")

        self.cmdLines = cmdLines
        self.masterCmd = self._writeCondorFiles()
        self.clusterId = None  # will be set by submit
        self._statuses = [0] * len(commandList)

    MasterCmd = (
        "should_transfer_files   = YES\n"
        "when_to_transfer_output = ON_EXIT\n"
        "success_exit_code = 0\n"
        "max_retries = 0\n"
        "arguments      = $(Process)\n"
        "executable     = {indir}/condor.sh\n"
        "output         = {logdir_rel}/condor_$(Process).out\n"
        "error          = {logdir_rel}/condor_$(Process).err\n"
        "log            = {logdir_rel}/condor_$(Process).log\n"
        "queue {nJobs:d}\n"
    )
    MasterShell = (
        "#!/usr/bin/env bash\n"
        "\n"
        ". {indir}/condor_$1.sh\n"
    )

    JobShell = (
        "#!/usr/bin/env bash\n"
        "\n"
        "{environment_setup}"
        "\n"
        "function move_files {{\n"
        "{move_fragment}"
        "\n}}\n"
        "\n"
        "{command} && move_files"
    )

    def _writeCondorFiles(self):
        """ Create Condor .sh and .cmd files """
        masterCmdName = os.path.join(self.workDirs["in"], "condor.cmd")
        with open(masterCmdName, "w") as masterCmd:
            if self.cmdLines:
                masterCmd.write("{}\n".format("\n".join(self.cmdLines)))
            masterCmd.write(CommandListJob.MasterCmd.format(
                indir=self.workDirs["in"],
                logdir_rel=os.path.relpath(self.workDirs["log"]),
                nJobs=len(self.commandList)
            ))
        masterShName = os.path.join(self.workDirs["in"], "condor.sh")
        with open(masterShName, "w") as masterSh:
            masterSh.write(CommandListJob.MasterShell.format(
                indir=self.workDirs["in"]
            ))
        makeExecutable(masterShName)

        for i, command in enumerate(self.commandList):
            jobShName = os.path.join(self.workDirs["in"], f"condor_{i:d}.sh")
            job_outdir = os.path.join(self.workDirs["out"], str(i))
            os.makedirs(job_outdir)
            with open(jobShName, "w") as jobSh:
                jobSh.write(CommandListJob.JobShell.format(
                    environment_setup="\n".join(self.envSetupLines),
                    move_fragment="\n".join((
                        " for file in {pattern}; do\n"
                        '   echo "Moving $file to {outdir}/"\n'
                        "   mv $file {outdir}/\n"
                        " done"
                    ).format(pattern=ipatt, outdir=job_outdir)
                        for ipatt in self.outputPatterns),
                    command=command
                ))
            makeExecutable(jobShName)

        return masterCmdName

    def _commandOutDir(self, command):
        """ Output directory for a given command """
        return os.path.join(self.workDirs["out"], str(self.commandList.index(command)))

    def commandOutFiles(self, command):
        """ Output files for a given command """
        import fnmatch
        cmdOutDir = self._commandOutDir(command)
        return list(os.path.join(cmdOutDir, fn) for fn in os.listdir(cmdOutDir)
                    if any(fnmatch.fnmatch(fn, pat) for pat in self.outputPatterns))

    def submit(self):
        """ Submit the jobs to condor """
        logger.info(f"Submitting {len(self.commandList):d} condor jobs.")
        self.clusterId = _submit(self.masterCmd)

        # save to file in case
        with open(os.path.join(self.workDirs["in"], "cluster_id"), "w") as f:
            f.write(self.clusterId)

        logger.info(f"Submitted, job ID is {self.clusterId}")

    def cancel(self):
        """ Cancel (condor_rm) """
        subprocess.check_call(["condor_rm", self.clusterId])

    def statuses(self, update=True):
        """ list of subjob statuses (numeric, using indices in CondorJobStatus) """
        if update:
            try:
                self.updateStatuses()
            except Exception as ex:
                logger.error(f"Exception while updating statuses (will reuse previous): {ex!s}")
        return [CondorJobStatus.index(sjst) for sjst in self._statuses]

    @property
    def status(self):
        statuses = self.statuses()
        if all(st == statuses[0] for st in statuses):
            return CondorJobStatus[statuses[0]]
        elif any(st == "RUNNING" for st in statuses):
            return "RUNNING"
        elif any(st == "REMOVED" for st in statuses):
            return "REMOVED"
        elif any(st == "FAILED" for st in statuses):
            return "FAILED"
        else:
            return "UNKNOWN"

    def subjobStatus(self, i):
        return self._statuses[i]

    def updateStatuses(self):
        if self.clusterId is None:
            raise Exception("Cannot get status before submitting the jobs to condor")
        res = list(self._statuses)
        qOut = subprocess.check_output(
            ["condor_q", self.clusterId, "-af", "ProcId", "JobStatus"]).decode().strip()
        if qOut:
            for sjLn in qOut.split("\n"):
                sjId, sjSt = sjLn.split()
                res[int(sjId)] = CondorJobStatus[int(sjSt)]
        hOut = subprocess.check_output(
            ["condor_history", self.clusterId, "-af", "ProcId", "JobStatus", "ExitCode"]).decode().strip()
        if hOut:
            for sjLn in hOut.split("\n"):
                sjId, sjSt, sjXc = sjLn.split()
                res[int(sjId)] = ("FAILED" if sjSt == "4" and sjXc != "0" else CondorJobStatus[int(sjSt)])
        self._statuses = res

    def commandStatus(self, command):
        return self.subjobStatus(self.commandList.index(command))

    def getID(self, command):
        return self.commandList.index(command)

    def getLogFile(self, command):
        return os.path.join(self.workDirs["log"], f"condor_{self.commandList.index(command)}.out")

    def getResubmitCommand(self, failedCommands):
        return getResubmitCommand(self.masterCmd, [self.commandList.index(cmd) for cmd in failedCommands])

    def getRuntime(self, command):
        chCmdArgs = [
            "condor_history", f"{self.clusterId}.{self.commandList.index(command):d}",
            "-af", "CommittedTime", "CommittedSuspensionTime"]
        elapsed, suspended = subprocess.check_output(chCmdArgs).decode().strip().split()
        import datetime
        return datetime.timedelta(seconds=float(elapsed) - float(suspended))


def jobsFromTasks(taskList, workdir=None, batchConfig=None, configOpts=None):
    cmdLines = []
    envSetupLines = []
    if batchConfig:
        if "requirements" in batchConfig:
            configOpts["cmd"]["requirements"] = batchConfig["requirements"]
        if "jobflavour" in batchConfig:
            configOpts["cmd"]["+JobFlavour"] = batchConfig["jobflavour"]
            configOpts["cmd"].pop("+MaxRuntime")  # job flavour overrides maxruntime
        if "maxruntime" in batchConfig:
            configOpts["cmd"]["+MaxRuntime"] = batchConfig["maxruntime"]
    if configOpts:
        cmdLines += [f"{key} = {value}" for key, value in configOpts.get("cmd", {}).items()]
        envSetupLines += configOpts.get("env", [])
    condorJob = CommandListJob(list(chain.from_iterable(task.commandList for task in taskList)),
                               workDir=workdir, cmdLines=cmdLines, envSetupLines=envSetupLines)
    for task in taskList:
        task.jobCluster = condorJob
    return [condorJob]


def makeTasksMonitor(jobs, tasks, interval=120):
    """ make a TasksMonitor for condor jobs """
    from .batch import TasksMonitor
    return TasksMonitor(
        jobs, tasks, interval=interval,
        allStatuses=CondorJobStatus,
        activeStatuses=(1, 2, 5),
        failedStatuses=(3, 6, 7),
        completedStatus=4
    )


def findOutputsForCommands(batchDir, commandMatchers):
    """
    Look for outputs of matching commands inside batch submission directory

    :param batchDir: batch submission directory (with an ``input/condor.cmd`` file)
    :param commandMatchers: a dictionary with matcher objects
        (return ``True`` when passed matching commands)

    :returns: tuple of a matches dictionary (same keys as commandMatchers,
        a list of output files from matching commands) and a list of IDs for subjobs without output
    """
    with open(os.path.join(batchDir, "input", "condor.cmd")) as cmdFile:
        nJobs = int(next(ln for ln in cmdFile if ln.startswith("queue ")).split()[1])
    cmds = []
    for iJ in range(nJobs):
        with open(os.path.join(batchDir, "input", f"condor_{iJ:d}.sh")) as jobShFile:
            cmdLine = next(ln for ln in jobShFile if ln.strip().endswith(" && move_files"))
            cmds.append(cmdLine.split(" && ")[0])
    matches = dict()
    id_noOut = []
    for mName, matcher in commandMatchers.items():
        ids_matched = [(i, cmd) for i, cmd in zip(count(), cmds) if matcher(cmd)]
        files_found = []
        if not ids_matched:
            logger.warning(f"No jobs matched for {mName}")
        else:
            for sjId, cmd in ids_matched:
                outdir = os.path.join(batchDir, "output", str(sjId))
                if not os.path.exists(outdir):
                    logger.debug(f"Output directory for {mName} not found: {outdir} (command: {cmd})")
                    id_noOut.append(sjId)
                else:
                    sjOut = [os.path.join(outdir, fn) for fn in os.listdir(outdir)]
                    if not sjOut:
                        logger.debug(f"No output files for {mName} found in {outdir} (command: {cmd})")
                        id_noOut.append(sjId)
                    files_found += sjOut
        matches[mName] = len(ids_matched), files_found
    return matches, sorted(id_noOut)


def _submit(cmdFile):
    # helper method, wrap condor_submit and return the job ID
    result = subprocess.check_output(["condor_submit", cmdFile]).decode()
    import re
    pat = re.compile(r"\d+ job\(s\) submitted to cluster (\d+)\.")
    g = pat.search(result)
    return g.group(1)


def getResubmitCommand(submissionScript, idsToResubmit):
    if os.path.isdir(submissionScript):
        submissionScript = os.path.join(submissionScript, "input", "condor.cmd")
    if not os.path.isfile(submissionScript):
        raise FileNotFoundError(submissionScript)
    return ["bambooHTCondorResubmit",
            "--ids={}".format(",".join(str(sjId) for sjId in idsToResubmit)),
            submissionScript]


def resubmit_cli(args=None):
    import argparse
    parser = argparse.ArgumentParser(description="Resubmit failed bamboo HTCondor jobs")
    parser.add_argument(
        "--ids", required=True,
        help="IDs of the subjobs to resubmit (comma-separated)")
    parser.add_argument(
        "--add", action="append",
        help="Line(s) to add to the file, after the original settings (can be passed multiple times)")
    parser.add_argument(
        "-n", "--nosubmit", action="store_true",
        help="Generate the resubmission command file, but stop short of submitting it")
    parser.add_argument(
        "origsubmit",
        help="Original condor submit file (typically '<output>/batch/input/condor.cmd')")
    params = parser.parse_args(args=args)

    logging.basicConfig(level=logging.INFO)

    if not os.path.isfile(params.origsubmit):
        raise FileNotFoundError(params.origsubmit)
    resubIds = params.ids.split(",")
    with open(params.origsubmit) as cdF:
        orig_lines = [ln.strip() for ln in cdF if ln.strip()]
    # find the 'arguments = " and "queue N" lines
    il_args, il_queue = None, None
    for i, ln in enumerate(orig_lines):
        if ln.startswith("arguments "):
            if il_args is not None:
                raise RuntimeError(
                    f"Multiple lines starting with 'arguments' in {params.origsubmit}. "
                    "This script should be called with the original submission script")
            il_args = i
        if ln.startswith("queue "):
            if il_queue is not None:
                raise RuntimeError(
                    f"Multiple lines starting with 'queue' in {params.origsubmit}. "
                    "This script should be called with the original submission script")
            il_queue = i
    if il_args is None or il_queue is None:
        raise RuntimeError(f"No arguments and queue lines found in {params.origsubmit}")
    assert il_args < il_queue
    if il_queue != len(orig_lines) - 1:
        logger.warning(f"The part of {params.origsubmit} beyond line {il_queue+1:d} will be ignored")
    resub_lines = (orig_lines[:il_args] + orig_lines[il_args + 1:il_queue]
                   + ([al for al in params.add] if params.add else [])
                   + list(chain.from_iterable(
                       [orig_lines[il_args].replace("$(Process)", sjId), "queue"]
                       for sjId in resubIds)))
    iresub = 1
    resubName = "{1}_{0:d}{2}".format(iresub, *os.path.splitext(params.origsubmit))
    while os.path.exists(resubName):
        iresub += 1
        resubName = "{1}_{0:d}.{2}".format(iresub, *os.path.splitext(params.origsubmit))
    with open(resubName, "w") as resubF:
        resubF.write("\n".join(resub_lines))
    if params.nosubmit:
        logger.info(
            f"Resubmission commands written to {resubName}, "
            "this can be submitted with 'condor_submit'")
    else:
        logger.info(f"Resubmitting {len(resubIds):d} condor jobs")
        clusterId = _submit(resubName)
        logger.info(f"Submitted, job ID is {clusterId}")
        with open(os.path.join(os.path.dirname(params.origsubmit), f"cluster_id_{iresub:d}"), "w") as clIdF:
            clIdF.write(clusterId)


def configCPUReq(batchOptDict, threads):
    batchOptDict["cmd"]["RequestCpus"] = threads



import os.path
import shutil
import subprocess

import pytest

pytestmark = pytest.mark.skipif(shutil.which("bambooRun") is None, reason="bambooRun was not found")

testData = os.path.join(os.path.dirname(__file__), "data")
examples = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "examples")


def test_bambooRun_help():
    assert subprocess.run(["bambooRun", "--help"]).returncode == 0


def test_bambooRun_help_nanozmm():
    assert subprocess.run([
        "bambooRun",
        "--module={}:NanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        "--help"]).returncode == 0


@pytest.mark.skipif(shutil.which("plotIt") is None, reason="plotIt was not found")
def test_bambooRun_nanozmm_1():
    assert subprocess.run([
        "bambooRun",
        "--module={}:NanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        os.path.join(examples, "test1.yml"),
        "--output=nanozmm_test1"]).returncode == 0


def test_bambooRun_nanozmm_worker():
    assert subprocess.run([
        "bambooRun",
        "--module={}:NanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        "--distributed=worker",
        "--sample=DY_M50_test",
        "--anaConfig={}".format(os.path.join(examples, "test1.yml")),
        os.path.join(testData, "DY_M50_2016.root"),
        "--output=test_worker_1.root"]).returncode == 0


def test_bambooRun_nanozmm_worker_debug():
    assert subprocess.run([
        "bambooRun",
        "--module={}:NanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        "--distributed=worker",
        "--backend=debug",
        "--sample=DY_M50_test",
        "--anaConfig={}".format(os.path.join(examples, "test1.yml")),
        os.path.join(testData, "DY_M50_2016.root"),
        "--output=test_worker_2.root"]).returncode == 0


@pytest.mark.withcompiled
def test_bambooRun_nanozmm_worker_compiled():
    assert subprocess.run([
        "bambooRun",
        "--module={}:NanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        "--distributed=worker",
        "--backend=compiled",
        "--sample=DY_M50_test",
        "--anaConfig={}".format(os.path.join(examples, "test1.yml")),
        os.path.join(testData, "DY_M50_2016.root"),
        "--output=test_worker_3.root"]).returncode == 0


def test_bambooRun_nanozmm_skimmer():
    assert subprocess.run([
        "bambooRun",
        "--module={}:SkimNanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        os.path.join(examples, "test1.yml"),
        "--output=nanozmm_test2.1"]).returncode == 0


def test_bambooRun_nanozmm_skimmer_debug():
    assert subprocess.run([
        "bambooRun",
        "--module={}:SkimNanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        "--backend=debug",
        os.path.join(examples, "test1.yml"),
        "--output=nanozmm_test2.2"]).returncode == 0


@pytest.mark.withcompiled
def test_bambooRun_nanozmm_skimmer_compiled():
    assert subprocess.run([
        "bambooRun",
        "--module={}:SkimNanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        "--backend=compiled",
        os.path.join(examples, "test1.yml"),
        "--output=nanozmm_test2.3"]).returncode == 0
